from poke_env.environment import (
    abstract_battle,
    battle,
    double_battle,
    effect,
    field,
    move,
    move_category,
    observation,
    observed_pokemon,
    pokemon,
    pokemon_gender,
    pokemon_type,
    side_condition,
    status,
    target,
    weather,
    z_crystal,
)
from poke_env.environment.abstract_battle import AbstractBattle
from poke_env.environment.battle import Battle
from poke_env.environment.double_battle import DoubleBattle
from poke_env.environment.effect import Effect
from poke_env.environment.field import Field
from poke_env.environment.move import SPECIAL_MOVES, EmptyMove, Move
from poke_env.environment.move_category import MoveCategory
from poke_env.environment.observation import Observation
from poke_env.environment.observed_pokemon import ObservedPokemon
from poke_env.environment.pokemon import Pokemon
from poke_env.environment.pokemon_gender import PokemonGender
from poke_env.environment.pokemon_type import PokemonType
from poke_env.environment.side_condition import STACKABLE_CONDITIONS, SideCondition
from poke_env.environment.status import Status
from poke_env.environment.target import Target
from poke_env.environment.weather import Weather
from poke_env.environment.z_crystal import Z_CRYSTAL

__all__ = [
    "AbstractBattle",
    "Battle",
    "DoubleBattle",
    "Effect",
    "EmptyMove",
    "Field",
    "Move",
    "MoveCategory",
    "Observation",
    "ObservedPokemon",
    "Pokemon",
    "PokemonGender",
    "PokemonType",
    "SPECIAL_MOVES",
    "STACKABLE_CONDITIONS",
    "SideCondition",
    "Status",
    "Target",
    "Weather",
    "Z_CRYSTAL",
    "abstract_battle",
    "battle",
    "double_battle",
    "effect",
    "field",
    "move",
    "move_category",
    "observation",
    "observed_pokemon",
    "pokemon",
    "pokemon_gender",
    "pokemon_type",
    "side_condition",
    "status",
    "target",
    "weather",
    "z_crystal",
]



from logging import Logger
from typing import Any, Dict, List, Optional, Union

from poke_env.environment.abstract_battle import AbstractBattle
from poke_env.environment.move import SPECIAL_MOVES, Move
from poke_env.environment.move_category import MoveCategory
from poke_env.environment.pokemon import Pokemon
from poke_env.environment.pokemon_type import PokemonType
from poke_env.environment.target import Target


class DoubleBattle(AbstractBattle):
    POKEMON_1_POSITION = -1
    POKEMON_2_POSITION = -2
    OPPONENT_1_POSITION = 1
    OPPONENT_2_POSITION = 2
    EMPTY_TARGET_POSITION = 0  # symbolic, not used by showdown

    def __init__(
        self,
        battle_tag: str,
        username: str,
        logger: Logger,
        gen: int,
        save_replays: Union[str, bool] = False,
    ):
        super(DoubleBattle, self).__init__(
            battle_tag, username, logger, save_replays, gen=gen
        )

        # Turn choice attributes
        self._available_moves: List[List[Move]] = [[], []]
        self._available_switches: List[List[Pokemon]] = [[], []]
        self._can_mega_evolve: List[bool] = [False, False]
        self._can_z_move: List[bool] = [False, False]
        self._can_dynamax: List[bool] = [False, False]
        self._can_tera: List[Union[bool, PokemonType]] = [False, False]
        self._opponent_can_dynamax: List[bool] = [True, True]
        self._opponent_can_mega_evolve: List[bool] = [True, True]
        self._opponent_can_z_move: List[bool] = [True, True]
        self._force_switch: List[bool] = [False, False]
        self._maybe_trapped: List[bool] = [False, False]
        self._trapped: List[bool] = [False, False]

        # Battle state attributes
        self._active_pokemon: Dict[str, Pokemon] = {}
        self._opponent_active_pokemon: Dict[str, Pokemon] = {}

        # Other
        self._move_to_pokemon_id: Dict[Move, str] = {}

    def clear_all_boosts(self):
        for active_pokemon_group in (self.active_pokemon, self.opponent_active_pokemon):
            for active_pokemon in active_pokemon_group:
                if active_pokemon is not None:
                    active_pokemon.clear_boosts()

    def end_illusion(self, pokemon_name: str, details: str):
        player_identifier = pokemon_name[:2]
        pokemon_identifier = pokemon_name[:3]
        if player_identifier == self._player_role:
            active_dict = self._active_pokemon
        else:
            active_dict = self._opponent_active_pokemon
        active = active_dict.get(pokemon_identifier)

        active_dict[pokemon_identifier] = self._end_illusion_on(
            illusioned=active, illusionist=pokemon_name, details=details
        )

    @staticmethod
    def _get_active_pokemon(
        active_pokemon: Dict[str, Pokemon], role: str
    ) -> List[Optional[Pokemon]]:
        pokemon_1 = active_pokemon.get(f"{role}a")
        pokemon_2 = active_pokemon.get(f"{role}b")
        if pokemon_1 is None or not pokemon_1.active or pokemon_1.fainted:
            pokemon_1 = None
        if pokemon_2 is None or not pokemon_2.active or pokemon_2.fainted:
            pokemon_2 = None
        return [pokemon_1, pokemon_2]

    def parse_request(self, request: Dict[str, Any]) -> None:
        """
        Update the object from a request.
        The player's pokemon are all updated, as well as available moves, switches and
        other related information (z move, mega evolution, forced switch...).

        :param request: Parsed JSON request object.
        :type request: dict
        """
        if self.logger is not None:
            self.logger.debug(
                "Parsing the following request update in battle %s:\n%s",
                self.battle_tag,
                request,
            )

        if "wait" in request and request["wait"]:
            self._wait = True
        else:
            self._wait = False

        self._available_moves = [[], []]
        self._available_switches = [[], []]
        self._can_mega_evolve = [False, False]
        self._can_z_move = [False, False]
        self._can_dynamax = [False, False]
        self._maybe_trapped = [False, False]
        self._trapped = [False, False]
        self._can_tera = [False, False]
        self._force_switch = request.get("forceSwitch", [False, False])
        self._reviving = any(
            [mon.get("reviving") for mon in request["side"]["pokemon"]]
        )

        if any(self._force_switch):
            self._move_on_next_request = True

        self._last_request = request

        if request.get("teamPreview", False):
            self._teampreview = True
            number_of_mons = len(request["side"]["pokemon"])
            self._max_team_size = request.get("maxTeamSize", number_of_mons)
        else:
            self._teampreview = False

        side = request["side"]
        if side["pokemon"]:
            self._player_role = side["pokemon"][0]["ident"][:2]
        self._update_team_from_request(side)

        if "active" in request:
            for active_pokemon_number, active_request in enumerate(request["active"]):
                pokemon_dict = request["side"]["pokemon"][active_pokemon_number]
                active_pokemon = self.get_pokemon(
                    pokemon_dict["ident"],
                    force_self_team=True,
                    details=pokemon_dict["details"],
                )
                if self.player_role is not None:
                    if (
                        active_pokemon_number == 0
                        and f"{self.player_role}a" not in self._active_pokemon
                    ):
                        self._active_pokemon[f"{self.player_role}a"] = active_pokemon
                    elif f"{self.player_role}b" not in self._active_pokemon:
                        self._active_pokemon[f"{self.player_role}b"] = active_pokemon
                    elif (
                        active_pokemon_number == 0
                        and self._active_pokemon[f"{self.player_role}a"].fainted
                        and self._active_pokemon[f"{self.player_role}b"]
                        == active_pokemon
                    ):
                        (
                            self._active_pokemon[f"{self.player_role}a"],
                            self._active_pokemon[f"{self.player_role}b"],
                        ) = (
                            self._active_pokemon[f"{self.player_role}b"],
                            self._active_pokemon[f"{self.player_role}a"],
                        )
                    elif (
                        active_pokemon_number == 1
                        and self._active_pokemon[f"{self.player_role}b"].fainted
                        and not active_pokemon.fainted
                        and self._active_pokemon[f"{self.player_role}a"]
                        == active_pokemon
                    ):
                        (
                            self._active_pokemon[f"{self.player_role}a"],
                            self._active_pokemon[f"{self.player_role}b"],
                        ) = (
                            self._active_pokemon[f"{self.player_role}b"],
                            self._active_pokemon[f"{self.player_role}a"],
                        )

                if active_pokemon.fainted:
                    continue

                if active_request.get("trapped"):
                    self._trapped[active_pokemon_number] = True

                self._available_moves[active_pokemon_number] = (
                    active_pokemon.available_moves_from_request(active_request)
                )

                if active_request.get("canMegaEvo", False):
                    self._can_mega_evolve[active_pokemon_number] = True
                if active_request.get("canZMove", False):
                    self._can_z_move[active_pokemon_number] = True
                if active_request.get("canDynamax", False):
                    self._can_dynamax[active_pokemon_number] = True
                if active_request.get("canTerastallize", False):
                    self._can_tera[active_pokemon_number] = PokemonType.from_name(
                        active_request["canTerastallize"]
                    )
                if active_request.get("maybeTrapped", False):
                    self._maybe_trapped[active_pokemon_number] = True

        for pokemon_index, trapped in enumerate(self.trapped):
            if (not trapped) or self.force_switch[pokemon_index]:
                for pokemon in side["pokemon"]:
                    if pokemon:
                        pokemon = self._team[pokemon["ident"]]
                        if (
                            not self.reviving
                            and not pokemon.active
                            and not pokemon.fainted
                        ):
                            self._available_switches[pokemon_index].append(pokemon)
                        if self.reviving and not pokemon.active and pokemon.fainted:
                            self._available_switches[pokemon_index].append(pokemon)

    def switch(self, pokemon_str: str, details: str, hp_status: str):
        pokemon_identifier = pokemon_str.split(":")[0][:3]
        player_identifier = pokemon_identifier[:2]
        team = (
            self._active_pokemon
            if player_identifier == self._player_role
            else self._opponent_active_pokemon
        )
        pokemon_out = team.pop(pokemon_identifier, None)
        if pokemon_out is not None:
            pokemon_out.switch_out()
        pokemon_in = self.get_pokemon(pokemon_str, details=details)
        pokemon_in.switch_in()
        pokemon_in.set_hp_status(hp_status)
        team[pokemon_identifier] = pokemon_in

    def _swap(self, pokemon_str: str, slot: str):
        player_identifier = pokemon_str.split(":")[0][:2]
        active = (
            self._active_pokemon
            if player_identifier == self.player_role
            else self._opponent_active_pokemon
        )
        slot_a = f"{player_identifier}a"
        slot_b = f"{player_identifier}b"

        if active[slot_a].fainted or active[slot_b].fainted:
            return

        slot_a_mon = active[slot_a]
        slot_b_mon = active[slot_b]

        pokemon = self.get_pokemon(pokemon_str)

        if (slot == "0" and pokemon == slot_a_mon) or (
            slot == "1" and pokemon == slot_b_mon
        ):
            pass
        else:
            active[slot_a], active[slot_b] = active[slot_b], active[slot_a]

    def get_possible_showdown_targets(
        self, move: Move, pokemon: Pokemon, dynamax: bool = False
    ) -> List[int]:
        """
        Given move of an ALLY Pokemon, returns a list of possible Pokemon Showdown
        targets for it. This is smart enough so that it figures whether the Pokemon
        is already dynamaxed.

        :param move: Move instance for which possible targets should be returned
        :type move: Move
        :param dynamax: whether given move also STARTS dynamax for its user
        :return: a list of integers indicating Pokemon Showdown targets:
            -1, -2, 1, 2 or self.EMPTY_TARGET_POSITION that indicates "no target"
        :rtype: List[int]
        """
        if move.id in SPECIAL_MOVES:
            return [self.EMPTY_TARGET_POSITION]

        pokemon_1, pokemon_2 = self.active_pokemon
        if pokemon == pokemon_1 and move in self.available_moves[0]:
            self_position = self.POKEMON_1_POSITION
            ally_position = self.POKEMON_2_POSITION
        elif pokemon == pokemon_2 and move in self.available_moves[1]:
            self_position = self.POKEMON_2_POSITION
            ally_position = self.POKEMON_1_POSITION
        else:
            raise Exception(
                f"Selected move {move.id} is not owned by any active ally Pokemon "
                f"that is currently battling"
            )

        if dynamax or pokemon.is_dynamaxed:
            if move.category == MoveCategory.STATUS:
                targets = [self.EMPTY_TARGET_POSITION]
            else:
                targets = [self.OPPONENT_1_POSITION, self.OPPONENT_2_POSITION]
        elif move.non_ghost_target and (
            PokemonType.GHOST not in pokemon.types
        ):  # fixing target for Curse
            return [self.EMPTY_TARGET_POSITION]
        elif move.id == "terastarstorm" and pokemon.type_1 == PokemonType.STELLAR:
            targets = [self.EMPTY_TARGET_POSITION]
        else:
            targets = {
                Target.from_showdown_message("adjacentAlly"): [ally_position],
                Target.from_showdown_message("adjacentAllyOrSelf"): [
                    ally_position,
                    self_position,
                ],
                Target.from_showdown_message("adjacentFoe"): [
                    self.OPPONENT_1_POSITION,
                    self.OPPONENT_2_POSITION,
                ],
                Target.from_showdown_message("all"): [self.EMPTY_TARGET_POSITION],
                Target.from_showdown_message("allAdjacent"): [
                    self.EMPTY_TARGET_POSITION
                ],
                Target.from_showdown_message("allAdjacentFoes"): [
                    self.EMPTY_TARGET_POSITION
                ],
                Target.from_showdown_message("allies"): [self.EMPTY_TARGET_POSITION],
                Target.from_showdown_message("allySide"): [self.EMPTY_TARGET_POSITION],
                Target.from_showdown_message("allyTeam"): [self.EMPTY_TARGET_POSITION],
                Target.from_showdown_message("any"): [
                    ally_position,
                    self.OPPONENT_1_POSITION,
                    self.OPPONENT_2_POSITION,
                ],
                Target.from_showdown_message("foeSide"): [self.EMPTY_TARGET_POSITION],
                Target.from_showdown_message("normal"): [
                    ally_position,
                    self.OPPONENT_1_POSITION,
                    self.OPPONENT_2_POSITION,
                ],
                Target.from_showdown_message("randomNormal"): [
                    self.EMPTY_TARGET_POSITION
                ],
                Target.from_showdown_message("scripted"): [self.EMPTY_TARGET_POSITION],
                Target.from_showdown_message("self"): [self.EMPTY_TARGET_POSITION],
                self.EMPTY_TARGET_POSITION: [self.EMPTY_TARGET_POSITION],
                None: [self.OPPONENT_1_POSITION, self.OPPONENT_2_POSITION],
            }[move.deduced_target]

        pokemon_ids = set(self._opponent_active_pokemon.keys())
        pokemon_ids.update(self._active_pokemon.keys())
        targets_to_keep = {
            {
                f"{self.player_role}a": -1,
                f"{self.player_role}b": -2,
                f"{self.opponent_role}a": 1,
                f"{self.opponent_role}b": 2,
            }[pokemon_identifier]
            for pokemon_identifier in pokemon_ids
        }
        targets_to_keep.add(self.EMPTY_TARGET_POSITION)
        targets = [target for target in targets if target in targets_to_keep]

        return targets

    def to_showdown_target(self, move: Move, target_mon: Optional[Pokemon]) -> int:
        """Returns the correct Showdown target of the Pokemon to be targeted.
        It will return 0 if no target is needed or if the target_mon is not
        an active pokemon; this is meaningless in showdown

        :param move: the move to be used against the target_mon
        :type move: Move
        :param target_mon: the Pokemon that is to be targeted
        :type target_mon: as implemented in poke-env
        :return: The corresponding showdown target if needed, otherwise 0
        :rtype: int
        """

        if (
            move.target
            and move.target in (Target.ANY, Target.NORMAL, Target.ADJACENT_FOE)
            and target_mon is not None
        ):
            if target_mon == self.active_pokemon[0]:
                return self.POKEMON_1_POSITION
            elif target_mon == self.active_pokemon[1]:
                return self.POKEMON_2_POSITION
            elif target_mon == self.opponent_active_pokemon[0]:
                return self.OPPONENT_1_POSITION
            else:
                return self.OPPONENT_2_POSITION

        # No need to return a target for each of the other target types
        else:
            return self.EMPTY_TARGET_POSITION

    @property
    def active_pokemon(self) -> List[Optional[Pokemon]]:
        """
        :return: The active pokemon, always at least one is not None
        :rtype: List[Optional[Pokemon]]
        """
        if self.player_role is None:
            raise ValueError("Unable to get active_pokemon, player_role is None")
        return self._get_active_pokemon(self._active_pokemon, self.player_role)

    @property
    def all_active_pokemons(self) -> List[Optional[Pokemon]]:
        """
        :return: A list containing all active pokemons and/or Nones.
        :rtype: List[Optional[Pokemon]]
        """
        return [*self.active_pokemon, *self.opponent_active_pokemon]

    @property
    def available_moves(self) -> List[List[Move]]:
        """
        :return: A list of two lists of moves the player can use during the current
            move request for each Pokemon.
        :rtype: List[List[Move]]
        """
        return self._available_moves

    @property
    def available_switches(self) -> List[List[Pokemon]]:
        """
        :return: The list of two lists of switches the player can do during the
            current move request for each active pokemon
        :rtype: List[List[Pokemon]]
        """
        return self._available_switches

    @property
    def can_mega_evolve(self) -> List[bool]:
        """
        :return: Whether or not either current active pokemon can mega evolve.
        :rtype: List[bool]
        """
        return self._can_mega_evolve

    @property
    def can_z_move(self) -> List[bool]:
        """
        :return: Whether or not the current active pokemon can z-move.
        :rtype: List[bool]
        """
        return self._can_z_move

    @property
    def can_dynamax(self) -> List[bool]:
        """
        :return: Whether or not the current active pokemon can dynamax
        :rtype: List[bool]
        """
        return self._can_dynamax

    @property
    def can_tera(self) -> List[Union[bool, PokemonType]]:
        """
        :return: Whether or not the current active pokemon can terastallize. If yes, will be a PokemonType.
        :rtype: List[Union[bool, PokemonType]]
        """
        return self._can_tera

    @property
    def force_switch(self) -> List[bool]:
        """
        :return: A boolean indicating whether the active pokemon is forced
            to switch out.
        :rtype: List[bool]
        """
        return self._force_switch

    @property
    def grounded(self) -> List[bool]:
        """
        :return: A boolean indicating whether the active pokemon are grounded
        :rtype: List[bool]
        """
        return [self.is_grounded(mon) if mon else True for mon in self.active_pokemon]

    @property
    def maybe_trapped(self) -> List[bool]:
        """
        :return: A boolean indicating whether either active pokemon is maybe trapped
            by the opponent.
        :rtype: List[bool]
        """
        return self._maybe_trapped

    @property
    def opponent_active_pokemon(self) -> List[Optional[Pokemon]]:
        """
        :return: The opponent active pokemon, always at least one is not None
        :rtype: List[Optional[Pokemon]]
        """
        if self.opponent_role is None:
            raise ValueError(
                "Unable to get opponent_active_pokemon, opponent_role is None"
            )
        return self._get_active_pokemon(
            self._opponent_active_pokemon, self.opponent_role
        )

    @property
    def opponent_can_dynamax(self) -> List[bool]:
        """
        :return: Whether or not opponent's current active pokemons can dynamax
        :rtype: List[bool]
        """
        return self._opponent_can_dynamax

    @opponent_can_dynamax.setter
    def opponent_can_dynamax(self, value: Union[bool, List[bool]]):
        if isinstance(value, bool):
            self._opponent_can_dynamax = [value, value]
        else:
            self._opponent_can_dynamax = value

    @property
    def opponent_can_mega_evolve(self) -> List[bool]:
        """
        :return: Whether or not opponent's current active pokemons can mega evolve
        :rtype: List[bool]
        """
        return self._opponent_can_mega_evolve

    @opponent_can_mega_evolve.setter
    def opponent_can_mega_evolve(self, value: Union[bool, List[bool]]):
        if isinstance(value, bool):
            self._opponent_can_mega_evolve = [value, value]
        else:
            self._opponent_can_mega_evolve = value  # type: ignore

    @property
    def opponent_can_z_move(self) -> List[bool]:
        """
        :return: Whether or not opponent's current active pokemons can z-move
        :rtype: List[bool]
        """
        return self._opponent_can_z_move

    @opponent_can_z_move.setter
    def opponent_can_z_move(self, value: Union[bool, List[bool]]):
        if isinstance(value, bool):
            self._opponent_can_z_move = [value, value]
        else:
            self._opponent_can_z_move = value  # type: ignore

    @property
    def trapped(self) -> List[bool]:
        """
        :return: A boolean indicating whether either active pokemon is trapped by the
            opponent.
        :rtype: List[bool]
        """
        return self._trapped

    @trapped.setter
    def trapped(self, value: List[bool]):
        self._trapped = value

    @property
    def reviving(self) -> bool:
        """
        :return: Whether or not any of the player's Pokemon is reviving.
        :rtype: bool
        """
        return self._reviving



"""This module defines the ObservedPokemon class, which stores
what we have observed about a pokemon throughout a battle, offering
a more efficient serialization of a pokemon
"""

import sys
from collections import OrderedDict
from copy import copy
from dataclasses import dataclass, field
from typing import Dict, List, Mapping, Optional, Union

from poke_env.environment.effect import Effect
from poke_env.environment.move import Move
from poke_env.environment.pokemon import Pokemon
from poke_env.environment.pokemon_gender import PokemonGender
from poke_env.environment.pokemon_type import PokemonType
from poke_env.environment.status import Status


@dataclass
class ObservedPokemon:
    species: str
    level: int
    name: str

    ability: Optional[str] = None
    boosts: Dict[str, int] = field(
        default_factory=lambda: {
            "accuracy": 0,
            "atk": 0,
            "def": 0,
            "evasion": 0,
            "spa": 0,
            "spd": 0,
            "spe": 0,
        }
    )
    current_hp_fraction: float = 1.0
    effects: Dict[Effect, int] = field(default_factory=dict)
    is_dynamaxed: bool = False
    is_terastallized: bool = False
    item: Optional[str] = None
    gender: Optional[PokemonGender] = None
    moves: "OrderedDict[str, Move]" = field(default_factory=OrderedDict)
    tera_type: Optional[PokemonType] = None
    shiny: Optional[bool] = None
    stats: Optional[Mapping[str, Union[List[int], int, None]]] = None
    status: Optional[Status] = None

    def to_pokemon(self, gen=9) -> Pokemon:
        mon = Pokemon(gen=gen, species=self.species, name=self.name)
        mon._item = self.item
        mon._level = self.level
        mon._moves = self.moves
        mon._ability = self.ability
        mon._terastallized_type = self.tera_type

        mon._stats = {}
        if self.stats:
            for stat in self.stats:
                if isinstance(self.stats[stat], int):
                    mon._stats[stat] = self.stats[stat]  # type: ignore

        mon._gender = self.gender
        mon._shiny = self.shiny
        return mon

    @staticmethod
    def initial_stats() -> Dict[str, List[int]]:
        return {
            "atk": [0, sys.maxsize],
            "def": [0, sys.maxsize],
            "spa": [0, sys.maxsize],
            "spd": [0, sys.maxsize],
            "spe": [0, sys.maxsize],
        }

    @staticmethod
    def from_pokemon(mon: Optional[Pokemon]):
        if mon is None:
            return None

        stats: Optional[Mapping[str, Union[List[int], int, None]]] = (
            ObservedPokemon.initial_stats()
        )
        if mon.stats is not None:
            stats = {k: v for (k, v) in mon.stats.items()}

        return ObservedPokemon(
            species=mon.species,
            level=mon.level,
            name=mon.name,
            ability=mon.ability,
            boosts={k: v for (k, v) in mon.boosts.items()},
            current_hp_fraction=mon.current_hp_fraction,
            effects={k: v for (k, v) in mon.effects.items()},
            is_dynamaxed=mon.is_dynamaxed,
            is_terastallized=mon.is_terastallized,
            item=mon.item,
            gender=mon.gender,
            moves=OrderedDict({k: copy(v) for (k, v) in mon.moves.items()}),
            tera_type=mon.tera_type,
            shiny=mon.shiny,
            stats=stats,
            status=mon.status,
        )



"""This module defines the Weather class, which represents a in-battle weather."""

import logging
from enum import Enum, auto


class Weather(Enum):
    """Enumeration, represent a non null weather in a battle."""

    UNKNOWN = auto()
    DESOLATELAND = auto()
    DELTASTREAM = auto()
    HAIL = auto()
    PRIMORDIALSEA = auto()
    RAINDANCE = auto()
    SANDSTORM = auto()
    SNOWSCAPE = SNOW = auto()
    SUNNYDAY = auto()

    def __str__(self) -> str:
        return f"{self.name} (weather) object"

    @staticmethod
    def from_showdown_message(message: str):
        """Returns the Weather object corresponding to the message.

        :param message: The message to convert.
        :type message: str
        :return: The corresponding Weather object.
        :rtype: Weather
        """
        message = message.replace("move: ", "")
        message = message.replace(" ", "_")
        message = message.replace("-", "_")

        try:
            return Weather[message.upper()]
        except KeyError:
            logging.getLogger("poke-env").warning(
                "Unexpected weather '%s' received. Weather.UNKNOWN will be used "
                "instead. If this is unexpected, please open an issue at "
                "https://github.com/hsahovic/poke-env/issues/ along with this error "
                "message and a description of your program.",
                message,
            )
            return Weather.UNKNOWN



"""This module defines the MoveCategory class, which represents a move category."""

from enum import Enum, auto, unique


@unique
class MoveCategory(Enum):
    """Enumeration, represent a move category."""

    PHYSICAL = auto()
    SPECIAL = auto()
    STATUS = auto()

    def __str__(self) -> str:
        return f"{self.name} (move category) object"



"""This module defines a player class with the Gymnasium API on the main thread.
For a black-box implementation consider using the module env_player.
"""

import asyncio
import time
from abc import abstractmethod
from typing import Any, Awaitable, Dict, Generic, List, Optional, Tuple, TypeVar, Union
from weakref import WeakKeyDictionary

from gymnasium.spaces import Space
from pettingzoo.utils.env import ParallelEnv  # type: ignore[import-untyped]

from poke_env.concurrency import POKE_LOOP, create_in_poke_loop
from poke_env.environment.abstract_battle import AbstractBattle
from poke_env.player.battle_order import (
    BattleOrder,
    DefaultBattleOrder,
    ForfeitBattleOrder,
)
from poke_env.player.player import Player
from poke_env.ps_client import AccountConfiguration
from poke_env.ps_client.server_configuration import (
    LocalhostServerConfiguration,
    ServerConfiguration,
)
from poke_env.teambuilder.teambuilder import Teambuilder

ItemType = TypeVar("ItemType")
ObsType = TypeVar("ObsType")
ActionType = TypeVar("ActionType")


class _AsyncQueue(Generic[ItemType]):
    def __init__(self, queue: asyncio.Queue[ItemType]):
        self.queue = queue

    async def async_get(self) -> ItemType:
        return await self.queue.get()

    def get(self) -> ItemType:
        res = asyncio.run_coroutine_threadsafe(self.async_get(), POKE_LOOP)
        return res.result()

    def race_get(self, *events: asyncio.Event) -> Optional[ItemType]:
        async def _race_get() -> Optional[ItemType]:
            get_task = asyncio.create_task(self.async_get())
            wait_tasks = [asyncio.create_task(e.wait()) for e in events]
            done, pending = await asyncio.wait(
                {get_task, *wait_tasks},
                return_when=asyncio.FIRST_COMPLETED,
            )
            for p in pending:
                p.cancel()
            if get_task in done:
                return get_task.result()
            else:
                return None

        res = asyncio.run_coroutine_threadsafe(_race_get(), POKE_LOOP)
        return res.result()

    async def async_put(self, item: ItemType):
        await self.queue.put(item)

    def put(self, item: ItemType):
        task = asyncio.run_coroutine_threadsafe(self.queue.put(item), POKE_LOOP)
        task.result()

    def empty(self):
        return self.queue.empty()

    def join(self):
        task = asyncio.run_coroutine_threadsafe(self.queue.join(), POKE_LOOP)
        task.result()

    async def async_join(self):
        await self.queue.join()


class _EnvPlayer(Player):
    battle_queue: _AsyncQueue[AbstractBattle]
    order_queue: _AsyncQueue[BattleOrder]

    def __init__(
        self,
        username: str,
        **kwargs: Any,
    ):
        self.__class__.__name__ = username
        super().__init__(**kwargs)
        self.__class__.__name__ = "_EnvPlayer"
        self.battle_queue = _AsyncQueue(create_in_poke_loop(asyncio.Queue, 1))
        self.order_queue = _AsyncQueue(create_in_poke_loop(asyncio.Queue, 1))
        self.battle: Optional[AbstractBattle] = None

    def choose_move(self, battle: AbstractBattle) -> Awaitable[BattleOrder]:
        return self._env_move(battle)

    async def _env_move(self, battle: AbstractBattle) -> BattleOrder:
        if not self.battle or self.battle.finished:
            self.battle = battle
        assert self.battle.battle_tag == battle.battle_tag
        await self.battle_queue.async_put(battle)
        order = await self.order_queue.async_get()
        return order

    def _battle_finished_callback(self, battle: AbstractBattle):
        asyncio.run_coroutine_threadsafe(self.battle_queue.async_put(battle), POKE_LOOP)


class PokeEnv(ParallelEnv[str, ObsType, ActionType]):
    """
    Base class implementing the Gymnasium API on the main thread.
    """

    _INIT_RETRIES = 100
    _TIME_BETWEEN_RETRIES = 0.5
    _SWITCH_CHALLENGE_TASK_RETRIES = 30
    _TIME_BETWEEN_SWITCH_RETRIES = 1

    def __init__(
        self,
        *,
        account_configuration1: Optional[AccountConfiguration] = None,
        account_configuration2: Optional[AccountConfiguration] = None,
        avatar: Optional[int] = None,
        battle_format: str = "gen8randombattle",
        log_level: Optional[int] = None,
        save_replays: Union[bool, str] = False,
        server_configuration: Optional[
            ServerConfiguration
        ] = LocalhostServerConfiguration,
        accept_open_team_sheet: Optional[bool] = False,
        start_timer_on_battle_start: bool = False,
        start_listening: bool = True,
        open_timeout: Optional[float] = 10.0,
        ping_interval: Optional[float] = 20.0,
        ping_timeout: Optional[float] = 20.0,
        team: Optional[Union[str, Teambuilder]] = None,
        start_challenging: bool = False,
        fake: bool = False,
        strict: bool = True,
    ):
        """
        :param account_configuration: Player configuration. If empty, defaults to an
            automatically generated username with no password. This option must be set
            if the server configuration requires authentication.
        :type account_configuration: AccountConfiguration, optional
        :param avatar: Player avatar id. Optional.
        :type avatar: int, optional
        :param battle_format: Name of the battle format this player plays. Defaults to
            gen8randombattle.
        :type battle_format: Optional, str. Default to randombattles, with specifics
            varying per class.
        :param log_level: The player's logger level.
        :type log_level: int. Defaults to logging's default level.
        :param save_replays: Whether to save battle replays. Can be a boolean, where
            True will lead to replays being saved in a potentially new /replay folder,
            or a string representing a folder where replays will be saved.
        :type save_replays: bool or str
        :param server_configuration: Server configuration. Defaults to Localhost Server
            Configuration.
        :type server_configuration: ServerConfiguration, optional
        :param start_listening: Whether to start listening to the server. Defaults to
            True.
        :type start_listening: bool
        :param accept_open_team_sheet: Whether to automatically start the battle with
            open team sheets on. Defaults to False.
        :param start_timer_on_battle_start: Whether to automatically start the battle
            timer on battle start. Defaults to False.
        :type start_timer_on_battle_start: bool
        :param open_timeout: How long to wait for a timeout when connecting the socket
            (important for backend websockets.
            Increase only if timeouts occur during runtime).
            If None connect will never time out.
        :type open_timeout: float, optional
        :param ping_interval: How long between keepalive pings (Important for backend
            websockets). If None, disables keepalive entirely.
        :type ping_interval: float, optional
        :param ping_timeout: How long to wait for a timeout of a specific ping
            (important for backend websockets.
            Increase only if timeouts occur during runtime).
            If None pings will never time out.
        :type ping_timeout: float, optional
        :param team: The team to use for formats requiring a team. Can be a showdown
            team string, a showdown packed team string, of a ShowdownTeam object.
            Defaults to None.
        :type team: str or Teambuilder, optional
        :param start_challenging: Whether to automatically start the challenge loop or
            leave it inactive.
        :type start_challenging: bool
        :param fake: If true, action-order converters will try to avoid returning a default
            output if at all possible, even if the output isn't a legal decision. Defaults
            to False.
        :type fake: bool
        :param strict: If true, action-order converters will throw an error if the move is
            illegal. Otherwise, it will return default. Defaults to True.
        :type: strict: bool
        """
        self.agent1 = _EnvPlayer(
            username=self.__class__.__name__,  # type: ignore
            account_configuration=account_configuration1,
            avatar=avatar,
            battle_format=battle_format,
            log_level=log_level,
            max_concurrent_battles=1,
            save_replays=save_replays,
            server_configuration=server_configuration,
            accept_open_team_sheet=accept_open_team_sheet,
            start_timer_on_battle_start=start_timer_on_battle_start,
            start_listening=start_listening,
            open_timeout=open_timeout,
            ping_interval=ping_interval,
            ping_timeout=ping_timeout,
            team=team,
        )
        self.agent2 = _EnvPlayer(
            username=self.__class__.__name__,  # type: ignore
            account_configuration=account_configuration2,
            avatar=avatar,
            battle_format=battle_format,
            log_level=log_level,
            max_concurrent_battles=1,
            save_replays=save_replays,
            server_configuration=server_configuration,
            accept_open_team_sheet=accept_open_team_sheet,
            start_timer_on_battle_start=start_timer_on_battle_start,
            start_listening=start_listening,
            open_timeout=open_timeout,
            ping_interval=ping_interval,
            ping_timeout=ping_timeout,
            team=team,
        )
        self.agents: List[str] = []
        self.possible_agents = [self.agent1.username, self.agent2.username]
        self.battle1: Optional[AbstractBattle] = None
        self.battle2: Optional[AbstractBattle] = None
        self.fake = fake
        self.strict = strict
        self._reward_buffer: WeakKeyDictionary[AbstractBattle, float] = (
            WeakKeyDictionary()
        )
        self._keep_challenging: bool = False
        self._challenge_task = None
        if start_challenging:
            self._keep_challenging = True
            self._challenge_task = asyncio.run_coroutine_threadsafe(
                self._challenge_loop(), POKE_LOOP
            )

    ###################################################################################
    # PettingZoo API
    # https://pettingzoo.farama.org/api/parallel/#parallelenv

    def step(self, actions: Dict[str, ActionType]) -> Tuple[
        Dict[str, ObsType],
        Dict[str, float],
        Dict[str, bool],
        Dict[str, bool],
        Dict[str, Dict[str, Any]],
    ]:
        assert self.battle1 is not None
        assert not self.battle1.finished
        assert self.battle2 is not None
        assert not self.battle2.finished
        agent1_waiting = self.agent1._waiting.is_set()
        agent2_waiting = self.agent2._waiting.is_set()
        agent1_trying_again = self.agent1._trying_again.is_set()
        agent2_trying_again = self.agent2._trying_again.is_set()
        self.agent1._waiting.clear()
        self.agent2._waiting.clear()
        self.agent1._trying_again.clear()
        self.agent2._trying_again.clear()
        if not (agent1_waiting or agent2_trying_again):
            order1 = self.action_to_order(
                actions[self.agents[0]],
                self.battle1,
                fake=self.fake,
                strict=self.strict,
            )
            self.agent1.order_queue.put(order1)
        if not (agent2_waiting or agent1_trying_again):
            order2 = self.action_to_order(
                actions[self.agents[1]],
                self.battle2,
                fake=self.fake,
                strict=self.strict,
            )
            self.agent2.order_queue.put(order2)
        battle1 = (
            self.agent1.battle_queue.race_get(
                self.agent1._waiting, self.agent2._trying_again
            )
            or self.battle1
        )
        battle2 = (
            self.agent2.battle_queue.race_get(
                self.agent2._waiting, self.agent1._trying_again
            )
            or self.battle2
        )
        observations = {
            self.agents[0]: self.embed_battle(battle1),
            self.agents[1]: self.embed_battle(battle2),
        }
        reward = {
            self.agents[0]: self.calc_reward(battle1),
            self.agents[1]: self.calc_reward(battle2),
        }
        term1, trunc1 = self.calc_term_trunc(battle1)
        term2, trunc2 = self.calc_term_trunc(battle2)
        terminated = {self.agents[0]: term1, self.agents[1]: term2}
        truncated = {self.agents[0]: trunc1, self.agents[1]: trunc2}
        if battle1.finished:
            self.agents = []
        return observations, reward, terminated, truncated, self.get_additional_info()

    def reset(
        self,
        seed: Optional[int] = None,
        options: Optional[Dict[str, Any]] = None,
    ) -> Tuple[Dict[str, ObsType], Dict[str, Dict[str, Any]]]:
        self.agents = [self.agent1.username, self.agent2.username]
        # TODO: use the seed
        if not self.agent1.battle or not self.agent2.battle:
            count = self._INIT_RETRIES
            while not self.agent1.battle or not self.agent2.battle:
                if count == 0:
                    raise RuntimeError("Agent is not challenging")
                count -= 1
                time.sleep(self._TIME_BETWEEN_RETRIES)
        if self.battle1 and not self.battle1.finished:
            assert self.battle2 is not None
            if self.battle1 == self.agent1.battle:
                agent1_waiting = self.agent1._waiting.is_set()
                agent2_waiting = self.agent2._waiting.is_set()
                agent1_trying_again = self.agent1._trying_again.is_set()
                agent2_trying_again = self.agent2._trying_again.is_set()
                self.agent1._waiting.clear()
                self.agent2._waiting.clear()
                self.agent1._trying_again.clear()
                self.agent2._trying_again.clear()
                if not (agent1_waiting or agent2_trying_again):
                    self.agent1.order_queue.put(ForfeitBattleOrder())
                    if not (agent2_waiting or agent1_trying_again):
                        self.agent2.order_queue.put(DefaultBattleOrder())
                else:
                    self.agent2.order_queue.put(ForfeitBattleOrder())
                self.agent1.battle_queue.get()
                self.agent2.battle_queue.get()
            else:
                raise RuntimeError(
                    "Environment and agent aren't synchronized. Try to restart"
                )
        self.battle1 = self.agent1.battle_queue.get()
        self.battle2 = self.agent2.battle_queue.get()
        observations = {
            self.agents[0]: self.embed_battle(self.battle1),
            self.agents[1]: self.embed_battle(self.battle2),
        }
        return observations, self.get_additional_info()

    def render(self, mode: str = "human"):
        if self.battle1 is not None:
            print(
                "  Turn %4d. | [%s][%3d/%3dhp] %10.10s - %10.10s [%3d%%hp][%s]"
                % (
                    self.battle1.turn,
                    "".join(
                        [
                            "⦻" if mon.fainted else "●"
                            for mon in self.battle1.team.values()
                        ]
                    ),
                    self.battle1.active_pokemon.current_hp or 0,
                    self.battle1.active_pokemon.max_hp or 0,
                    self.battle1.active_pokemon.species,
                    self.battle1.opponent_active_pokemon.species,
                    self.battle1.opponent_active_pokemon.current_hp or 0,
                    "".join(
                        [
                            "⦻" if mon.fainted else "●"
                            for mon in self.battle1.opponent_team.values()
                        ]
                    ),
                ),
                end="\n" if self.battle1.finished else "\r",
            )

    def close(self, purge: bool = True):
        if self.battle1 is None or self.battle1.finished:
            time.sleep(1)
            if self.battle1 != self.agent1.battle:
                self.battle1 = self.agent1.battle
        if self.battle2 is None or self.battle2.finished:
            time.sleep(1)
            if self.battle2 != self.agent2.battle:
                self.battle2 = self.agent2.battle
        closing_task = asyncio.run_coroutine_threadsafe(
            self._stop_challenge_loop(purge=purge), POKE_LOOP
        )
        closing_task.result()

    def observation_space(self, agent: str) -> Space[ObsType]:
        return self.observation_spaces[agent]

    def action_space(self, agent: str) -> Space[ActionType]:
        return self.action_spaces[agent]

    ###################################################################################
    # Abstract methods

    @abstractmethod
    def calc_reward(self, battle: AbstractBattle) -> float:
        """
        Returns the reward for the current battle state.

        :param battle: The current battle state.
        :type battle: AbstractBattle

        :return: The reward for battle.
        :rtype: float
        """
        pass

    @abstractmethod
    def embed_battle(self, battle: AbstractBattle) -> ObsType:
        """
        Returns the embedding of the current battle state in a format compatible with
        the Gymnasium API.

        :param battle: The current battle state.
        :type battle: AbstractBattle

        :return: The embedding of the current battle state.
        """
        pass

    @staticmethod
    @abstractmethod
    def action_to_order(
        action: ActionType, battle: Any, fake: bool = False, strict: bool = True
    ) -> BattleOrder:
        """
        Returns the BattleOrder relative to the given action.

        :param action: The action to take.
        :type action: ActionType
        :param battle: The current battle state
        :type battle: AbstractBattle

        :return: The battle order for the given action in context of the current battle.
        :rtype: BattleOrder
        """
        pass

    @staticmethod
    @abstractmethod
    def order_to_action(
        order: BattleOrder, battle: Any, fake: bool = False, strict: bool = True
    ) -> ActionType:
        """
        Returns the action relative to the given BattleOrder.

        :param order: The order to take.
        :type order: BattleOrder
        :param battle: The current battle state
        :type battle: AbstractBattle

        :return: The action for the given battle order in context of the current battle.
        :rtype: ActionType
        """
        pass

    ###################################################################################
    # Helper methods

    def reward_computing_helper(
        self,
        battle: AbstractBattle,
        *,
        fainted_value: float = 0.0,
        hp_value: float = 0.0,
        number_of_pokemons: int = 6,
        starting_value: float = 0.0,
        status_value: float = 0.0,
        victory_value: float = 1.0,
    ) -> float:
        """A helper function to compute rewards.

        The reward is computed by computing the value of a game state, and by comparing
        it to the last state.

        State values are computed by weighting different factor. Fainted pokemons,
        their remaining HP, inflicted statuses and winning are taken into account.

        For instance, if the last time this function was called for battle A it had
        a state value of 8 and this call leads to a value of 9, the returned reward will
        be 9 - 8 = 1.

        Consider a single battle where each player has 6 pokemons. No opponent pokemon
        has fainted, but our team has one fainted pokemon. Three opposing pokemons are
        burned. We have one pokemon missing half of its HP, and our fainted pokemon has
        no HP left.

        The value of this state will be:

        - With fainted value: 1, status value: 0.5, hp value: 1:
            = - 1 (fainted) + 3 * 0.5 (status) - 1.5 (our hp) = -1
        - With fainted value: 3, status value: 0, hp value: 1:
            = - 3 + 3 * 0 - 1.5 = -4.5

        :param battle: The battle for which to compute rewards.
        :type battle: AbstractBattle
        :param fainted_value: The reward weight for fainted pokemons. Defaults to 0.
        :type fainted_value: float
        :param hp_value: The reward weight for hp per pokemon. Defaults to 0.
        :type hp_value: float
        :param number_of_pokemons: The number of pokemons per team. Defaults to 6.
        :type number_of_pokemons: int
        :param starting_value: The default reference value evaluation. Defaults to 0.
        :type starting_value: float
        :param status_value: The reward value per non-fainted status. Defaults to 0.
        :type status_value: float
        :param victory_value: The reward value for winning. Defaults to 1.
        :type victory_value: float
        :return: The reward.
        :rtype: float
        """
        if battle not in self._reward_buffer:
            self._reward_buffer[battle] = starting_value
        current_value = 0.0

        for mon in battle.team.values():
            current_value += mon.current_hp_fraction * hp_value
            if mon.fainted:
                current_value -= fainted_value
            elif mon.status is not None:
                current_value -= status_value

        current_value += (number_of_pokemons - len(battle.team)) * hp_value

        for mon in battle.opponent_team.values():
            current_value -= mon.current_hp_fraction * hp_value
            if mon.fainted:
                current_value += fainted_value
            elif mon.status is not None:
                current_value += status_value

        current_value -= (number_of_pokemons - len(battle.opponent_team)) * hp_value

        if battle.won:
            current_value += victory_value
        elif battle.lost:
            current_value -= victory_value

        to_return = current_value - self._reward_buffer[battle]
        self._reward_buffer[battle] = current_value

        return to_return

    def reset_env(self, restart: bool = True):
        """
        Resets the environment to an inactive state: it will forfeit all unfinished
        battles, reset the internal battle tracker and optionally change the next
        opponent and restart the challenge loop.

        :param opponent: The opponent to use for the next battles. If empty it
            will not change opponent.
        :type opponent: Player or str, optional
        :param restart: If True the challenge loop will be restarted before returning,
            otherwise the challenge loop will be left inactive and can be
            started manually.
        :type restart: bool
        """
        self.close(purge=False)
        self.reset_battles()
        if restart:
            self.start_challenging()

    def get_additional_info(self) -> Dict[str, Dict[str, Any]]:
        """
        Returns additional info for the reset method.
        Override only if you really need it.

        :return: Additional information as a Dict
        :rtype: Dict
        """
        return {self.possible_agents[0]: {}, self.possible_agents[1]: {}}

    @staticmethod
    def calc_term_trunc(battle: AbstractBattle):
        terminated = False
        truncated = False
        if battle.finished:
            size = battle.team_size
            remaining_mons = size - len(
                [mon for mon in battle.team.values() if mon.fainted]
            )
            remaining_opponent_mons = size - len(
                [mon for mon in battle.opponent_team.values() if mon.fainted]
            )
            if (remaining_mons == 0) != (remaining_opponent_mons == 0):
                terminated = True
            else:
                truncated = True
        return terminated, truncated

    def background_send_challenge(self, username: str):
        """
        Sends a single challenge specified player. The function immediately returns
        to allow use of the Gymnasium API.

        :param username: The username of the player to challenge.
        :type username: str
        """
        if self._challenge_task and not self._challenge_task.done():
            raise RuntimeError(
                "Agent is already challenging opponents with the challenging loop. "
                "Try to specify 'start_challenging=True' during instantiation or call "
                "'await agent.stop_challenge_loop()' to clear the task."
            )
        self._challenge_task = asyncio.run_coroutine_threadsafe(
            self.agent1.send_challenges(username, 1), POKE_LOOP
        )

    def background_accept_challenge(self, username: str):
        """
        Accepts a single challenge specified player. The function immediately returns
        to allow use of the Gymnasium API.

        :param username: The username of the player to challenge.
        :type username: str
        """
        if self._challenge_task and not self._challenge_task.done():
            raise RuntimeError(
                "Agent is already challenging opponents with the challenging loop. "
                "Try to specify 'start_challenging=True' during instantiation or call "
                "'await agent.stop_challenge_loop()' to clear the task."
            )
        self._challenge_task = asyncio.run_coroutine_threadsafe(
            self.agent1.accept_challenges(username, 1, self.agent1.next_team), POKE_LOOP
        )

    async def _challenge_loop(self, n_challenges: Optional[int] = None):
        if not n_challenges:
            while self._keep_challenging:
                await self.agent1.battle_against(self.agent2, n_battles=1)
        elif n_challenges > 0:
            for _ in range(n_challenges):
                await self.agent1.battle_against(self.agent2, n_battles=1)
        else:
            raise ValueError(f"Number of challenges must be > 0. Got {n_challenges}")

    def start_challenging(self, n_challenges: Optional[int] = None):
        """
        Starts the challenge loop.

        :param n_challenges: The number of challenges to send. If empty it will run until
            stopped.
        :type n_challenges: int, optional
        """
        if self._challenge_task and not self._challenge_task.done():
            count = self._SWITCH_CHALLENGE_TASK_RETRIES
            while not self._challenge_task.done():
                if count == 0:
                    raise RuntimeError("Agent is already challenging")
                count -= 1
                time.sleep(self._TIME_BETWEEN_SWITCH_RETRIES)
        if not n_challenges:
            self._keep_challenging = True
        self._challenge_task = asyncio.run_coroutine_threadsafe(
            self._challenge_loop(n_challenges), POKE_LOOP
        )

    async def _ladder_loop(self, n_challenges: Optional[int] = None):
        if n_challenges:
            if n_challenges <= 0:
                raise ValueError(
                    f"Number of challenges must be > 0. Got {n_challenges}"
                )
            for _ in range(n_challenges):
                await self.agent1.ladder(1)
        else:
            while self._keep_challenging:
                await self.agent1.ladder(1)

    def start_laddering(self, n_challenges: Optional[int] = None):
        """
        Starts the laddering loop.

        :param n_challenges: The number of ladder games to play. If empty it
            will run until stopped.
        :type n_challenges: int, optional
        """
        if self._challenge_task and not self._challenge_task.done():
            count = self._SWITCH_CHALLENGE_TASK_RETRIES
            while not self._challenge_task.done():
                if count == 0:
                    raise RuntimeError("Agent is already challenging")
                count -= 1
                time.sleep(self._TIME_BETWEEN_SWITCH_RETRIES)
        if not n_challenges:
            self._keep_challenging = True
        self._challenge_task = asyncio.run_coroutine_threadsafe(
            self._ladder_loop(n_challenges), POKE_LOOP
        )

    async def _stop_challenge_loop(
        self, force: bool = True, wait: bool = True, purge: bool = False
    ):
        self._keep_challenging = False

        if force:
            if self.battle1 and not self.battle1.finished:
                assert self.battle2 is not None
                if not (
                    self.agent1.order_queue.empty() and self.agent2.order_queue.empty()
                ):
                    await asyncio.sleep(2)
                    if not (
                        self.agent1.order_queue.empty()
                        and self.agent2.order_queue.empty()
                    ):
                        raise RuntimeError(
                            "The agent is still sending actions. "
                            "Use this method only when training or "
                            "evaluation are over."
                        )
                if not self.agent1.battle_queue.empty():
                    await self.agent1.battle_queue.async_get()
                if not self.agent2.battle_queue.empty():
                    await self.agent2.battle_queue.async_get()
                agent1_waiting = self.agent1._waiting.is_set()
                agent2_waiting = self.agent2._waiting.is_set()
                agent1_trying_again = self.agent1._trying_again.is_set()
                agent2_trying_again = self.agent2._trying_again.is_set()
                self.agent1._waiting.clear()
                self.agent2._waiting.clear()
                self.agent1._trying_again.clear()
                self.agent2._trying_again.clear()
                if not (agent1_waiting or agent2_trying_again):
                    await self.agent1.order_queue.async_put(ForfeitBattleOrder())
                    if not (agent2_waiting or agent1_trying_again):
                        await self.agent2.order_queue.async_put(DefaultBattleOrder())
                else:
                    await self.agent2.order_queue.async_put(ForfeitBattleOrder())

        if wait and self._challenge_task:
            while not self._challenge_task.done():
                await asyncio.sleep(1)
            self._challenge_task.result()

        self._challenge_task = None
        self.battle1 = None
        self.battle2 = None
        self.agent1.battle = None
        self.agent2.battle = None
        while not self.agent1.order_queue.empty():
            await self.agent1.order_queue.async_get()
        while not self.agent2.order_queue.empty():
            await self.agent2.order_queue.async_get()
        while not self.agent1.battle_queue.empty():
            await self.agent1.battle_queue.async_get()
        while not self.agent2.battle_queue.empty():
            await self.agent2.battle_queue.async_get()

        if purge:
            self.reset_battles()

    def reset_battles(self):
        """Resets the player's inner battle tracker."""
        self.agent1.reset_battles()
        self.agent2.reset_battles()

    def done(self, timeout: Optional[int] = None) -> bool:
        """
        Returns True if the task is done or is done after the timeout, false otherwise.

        :param timeout: The amount of time to wait for if the task is not already done.
            If empty it will wait until the task is done.
        :type timeout: int, optional

        :return: True if the task is done or if the task gets completed after the
            timeout.
        :rtype: bool
        """
        if self._challenge_task is None:
            return True
        if timeout is None:
            self._challenge_task.result()
            return True
        if self._challenge_task.done():
            return True
        time.sleep(timeout)
        return self._challenge_task.done()



from __future__ import annotations

import os
from functools import lru_cache
from typing import Any, Dict, Optional, Union

import orjson

from poke_env.data.normalize import to_id_str


class GenData:
    __slots__ = ("gen", "moves", "natures", "pokedex", "type_chart", "learnset")

    UNKNOWN_ITEM = "unknown_item"

    _gen_data_per_gen: Dict[int, GenData] = {}

    def __init__(self, gen: int):
        if gen in self._gen_data_per_gen:
            raise ValueError(f"GenData for gen {gen} already initialized.")

        self.gen = gen
        self.moves = self.load_moves(gen)
        self.natures = self.load_natures()
        self.pokedex = self.load_pokedex(gen)
        self.type_chart = self.load_type_chart(gen)
        self.learnset = self.load_learnset()

    def __deepcopy__(self, memodict: Optional[Dict[int, Any]] = None) -> GenData:
        return self

    def load_moves(self, gen: int) -> Dict[str, Any]:
        with open(
            os.path.join(self._static_files_root, "moves", f"gen{gen}moves.json")
        ) as f:
            return orjson.loads(f.read())

    def load_natures(self) -> Dict[str, Dict[str, Union[int, float]]]:
        with open(os.path.join(self._static_files_root, "natures.json")) as f:
            return orjson.loads(f.read())

    def load_learnset(self) -> Dict[str, Dict[str, Union[int, float]]]:
        with open(os.path.join(self._static_files_root, "learnset.json")) as f:
            return orjson.loads(f.read())

    def load_pokedex(self, gen: int) -> Dict[str, Any]:
        with open(
            os.path.join(self._static_files_root, "pokedex", f"gen{gen}pokedex.json")
        ) as f:
            dex = orjson.loads(f.read())

        other_forms_dex: Dict[str, Any] = {}
        for value in dex.values():
            if "cosmeticFormes" in value:
                for other_form in value["cosmeticFormes"]:
                    other_forms_dex[to_id_str(other_form)] = value

        # Alternative pikachu gmax forms
        for name, value in dex.items():
            if name.startswith("pikachu") and name not in {"pikachu", "pikachugmax"}:
                other_forms_dex[name + "gmax"] = dex["pikachugmax"]

        dex.update(other_forms_dex)

        for name, value in dex.items():
            if "baseSpecies" in value:
                value["species"] = value["baseSpecies"]
            else:
                value["baseSpecies"] = to_id_str(name)

        return dex

    def load_type_chart(self, gen: int) -> Dict[str, Dict[str, float]]:
        with open(
            os.path.join(
                self._static_files_root, "typechart", f"gen{gen}typechart.json"
            )
        ) as chart:
            json_chart = orjson.loads(chart.read())

        types = [str(type_).upper() for type_ in json_chart]
        type_chart = {type_1: {type_2: 1.0 for type_2 in types} for type_1 in types}

        for type_, data in json_chart.items():
            type_ = type_.upper()

            for other_type, damage_taken in data["damageTaken"].items():
                if other_type.upper() not in types:
                    continue

                assert damage_taken in {0, 1, 2, 3}, (data["damageTaken"], type_)

                if damage_taken == 0:
                    type_chart[type_][other_type.upper()] = 1
                elif damage_taken == 1:
                    type_chart[type_][other_type.upper()] = 2
                elif damage_taken == 2:
                    type_chart[type_][other_type.upper()] = 0.5
                elif damage_taken == 3:
                    type_chart[type_][other_type.upper()] = 0

            assert set(types).issubset(set(type_chart))

        assert len(type_chart) == len(types)

        for effectiveness in type_chart.values():
            assert len(effectiveness) == len(types)

        return type_chart

    @property
    def _static_files_root(self) -> str:
        return os.path.join(os.path.dirname(os.path.realpath(__file__)), "static")

    @classmethod
    @lru_cache(None)
    def from_gen(cls, gen: int) -> GenData:
        gen_data = GenData(gen)
        cls._gen_data_per_gen[gen] = gen_data

        return gen_data

    @classmethod
    @lru_cache(None)
    def from_format(cls, format: str) -> GenData:
        gen = int(format[3])  # Update when Gen 10 comes
        return cls.from_gen(gen)



"""poke_env.player module init."""

from pettingzoo.utils.env import ActionType, ObsType  # type: ignore[import-untyped]

from poke_env.concurrency import POKE_LOOP
from poke_env.player import env, player, random_player, utils
from poke_env.player.baselines import MaxBasePowerPlayer, SimpleHeuristicsPlayer
from poke_env.player.battle_order import (
    BattleOrder,
    DefaultBattleOrder,
    DoubleBattleOrder,
    ForfeitBattleOrder,
)
from poke_env.player.env import PokeEnv
from poke_env.player.player import Player
from poke_env.player.random_player import RandomPlayer
from poke_env.player.singles_env import SinglesEnv
from poke_env.player.utils import (
    background_cross_evaluate,
    background_evaluate_player,
    cross_evaluate,
    evaluate_player,
)
from poke_env.ps_client import PSClient

__all__ = [
    "env",
    "player",
    "random_player",
    "utils",
    "ActionType",
    "ObsType",
    "ForfeitBattleOrder",
    "POKE_LOOP",
    "PokeEnv",
    "SinglesEnv",
    "PSClient",
    "Player",
    "RandomPlayer",
    "cross_evaluate",
    "background_cross_evaluate",
    "background_evaluate_player",
    "evaluate_player",
    "BattleOrder",
    "DefaultBattleOrder",
    "DoubleBattleOrder",
    "MaxBasePowerPlayer",
    "SimpleHeuristicsPlayer",
]



import random
from typing import List, Optional

from poke_env.environment.abstract_battle import AbstractBattle
from poke_env.environment.double_battle import DoubleBattle
from poke_env.environment.move_category import MoveCategory
from poke_env.environment.pokemon import Pokemon
from poke_env.environment.side_condition import SideCondition
from poke_env.environment.target import Target
from poke_env.player.battle_order import (
    BattleOrder,
    DefaultBattleOrder,
    DoubleBattleOrder,
)
from poke_env.player.player import Player


class MaxBasePowerPlayer(Player):
    def choose_move(self, battle: AbstractBattle):
        if self.format_is_doubles:
            return self.choose_doubles_move(battle)  # type: ignore
        else:
            return self.choose_singles_move(battle)

    def choose_singles_move(self, battle: AbstractBattle):
        if battle.available_moves:
            best_move = max(battle.available_moves, key=lambda move: move.base_power)
            return self.create_order(best_move)
        return self.choose_random_move(battle)

    def choose_doubles_move(self, battle: DoubleBattle):
        orders: List[Optional[BattleOrder]] = []
        switched_in = None

        if any(battle.force_switch):
            return self.choose_random_doubles_move(battle)

        can_target_first_opponent = (
            battle.opponent_active_pokemon[0]
            and not battle.opponent_active_pokemon[0].fainted
        )
        can_target_second_opponent = (
            battle.opponent_active_pokemon[1]
            and not battle.opponent_active_pokemon[1].fainted
        )
        can_double_target = can_target_first_opponent and can_target_second_opponent

        for mon, moves, switches in zip(
            battle.active_pokemon, battle.available_moves, battle.available_switches
        ):
            switches = [s for s in switches if s != switched_in]

            if not mon or mon.fainted:
                orders.append(None)
                continue
            elif not moves and switches:
                mon_to_switch_in = random.choice(switches)
                orders.append(BattleOrder(mon_to_switch_in))
                switched_in = mon_to_switch_in
                continue
            elif not moves:
                orders.append(DefaultBattleOrder())
                continue

            def move_power_with_double_target(move):
                if move.target in {Target.NORMAL, Target.ANY} or not can_double_target:
                    return move.base_power
                return move.base_power * 1.5

            best_move = max(moves, key=move_power_with_double_target)

            # randomly picks between the two opponents for normal move targeting
            targets = battle.get_possible_showdown_targets(best_move, mon)
            opp_targets = [
                t
                for t in targets
                if t in {battle.OPPONENT_1_POSITION, battle.OPPONENT_2_POSITION}
            ]
            if opp_targets:
                target = random.choice(opp_targets)
            else:
                target = random.choice(targets)

            orders.append(BattleOrder(best_move, move_target=target))

        if orders[0] or orders[1]:
            return DoubleBattleOrder(orders[0], orders[1])

        return self.choose_random_move(battle)


class SimpleHeuristicsPlayer(Player):
    ENTRY_HAZARDS = {
        "spikes": SideCondition.SPIKES,
        "stealhrock": SideCondition.STEALTH_ROCK,
        "stickyweb": SideCondition.STICKY_WEB,
        "toxicspikes": SideCondition.TOXIC_SPIKES,
    }

    ANTI_HAZARDS_MOVES = {"rapidspin", "defog"}

    SPEED_TIER_COEFICIENT = 0.1
    HP_FRACTION_COEFICIENT = 0.4
    SWITCH_OUT_MATCHUP_THRESHOLD = -2

    def _estimate_matchup(self, mon: Pokemon, opponent: Pokemon):
        score = max([opponent.damage_multiplier(t) for t in mon.types if t is not None])
        score -= max(
            [mon.damage_multiplier(t) for t in opponent.types if t is not None]
        )
        if mon.base_stats["spe"] > opponent.base_stats["spe"]:
            score += self.SPEED_TIER_COEFICIENT
        elif opponent.base_stats["spe"] > mon.base_stats["spe"]:
            score -= self.SPEED_TIER_COEFICIENT

        score += mon.current_hp_fraction * self.HP_FRACTION_COEFICIENT
        score -= opponent.current_hp_fraction * self.HP_FRACTION_COEFICIENT

        return score

    def _should_dynamax(self, battle: AbstractBattle, n_remaining_mons: int):
        if battle.can_dynamax:
            # Last full HP mon
            if (
                len([m for m in battle.team.values() if m.current_hp_fraction == 1])
                == 1
                and battle.active_pokemon.current_hp_fraction == 1
            ):
                return True
            # Matchup advantage and full hp on full hp
            if (
                self._estimate_matchup(
                    battle.active_pokemon, battle.opponent_active_pokemon
                )
                > 0
                and battle.active_pokemon.current_hp_fraction == 1
                and battle.opponent_active_pokemon.current_hp_fraction == 1
            ):
                return True
            if n_remaining_mons == 1:
                return True
        return False

    def _should_switch_out(self, battle: AbstractBattle):
        active = battle.active_pokemon
        opponent = battle.opponent_active_pokemon
        # If there is a decent switch in...
        if [
            m
            for m in battle.available_switches
            if self._estimate_matchup(m, opponent) > 0
        ]:
            # ...and a 'good' reason to switch out
            if active.boosts["def"] <= -3 or active.boosts["spd"] <= -3:
                return True
            if (
                active.boosts["atk"] <= -3
                and active.stats["atk"] >= active.stats["spa"]
            ):
                return True
            if (
                active.boosts["spa"] <= -3
                and active.stats["atk"] <= active.stats["spa"]
            ):
                return True
            if (
                self._estimate_matchup(active, opponent)
                < self.SWITCH_OUT_MATCHUP_THRESHOLD
            ):
                return True
        return False

    def _stat_estimation(self, mon: Pokemon, stat: str):
        # Stats boosts value
        if mon.boosts[stat] > 1:
            boost = (2 + mon.boosts[stat]) / 2
        else:
            boost = 2 / (2 - mon.boosts[stat])
        return ((2 * mon.base_stats[stat] + 31) + 5) * boost

    def choose_move(self, battle: AbstractBattle):
        if isinstance(battle, DoubleBattle):
            return self.choose_random_doubles_move(battle)

        # Main mons shortcuts
        active = battle.active_pokemon
        opponent = battle.opponent_active_pokemon

        if active is None or opponent is None:
            return self.choose_random_move(battle)

        # Rough estimation of damage ratio
        physical_ratio = self._stat_estimation(active, "atk") / self._stat_estimation(
            opponent, "def"
        )
        special_ratio = self._stat_estimation(active, "spa") / self._stat_estimation(
            opponent, "spd"
        )

        if battle.available_moves and (
            not self._should_switch_out(battle) or not battle.available_switches
        ):
            n_remaining_mons = len(
                [m for m in battle.team.values() if m.fainted is False]
            )
            n_opp_remaining_mons = 6 - len(
                [m for m in battle.opponent_team.values() if m.fainted is True]
            )

            # Entry hazard...
            for move in battle.available_moves:
                # ...setup
                if (
                    n_opp_remaining_mons >= 3
                    and move.id in self.ENTRY_HAZARDS
                    and self.ENTRY_HAZARDS[move.id]
                    not in battle.opponent_side_conditions
                ):
                    return self.create_order(move)

                # ...removal
                elif (
                    battle.side_conditions
                    and move.id in self.ANTI_HAZARDS_MOVES
                    and n_remaining_mons >= 2
                ):
                    return self.create_order(move)

            # Setup moves
            if (
                active.current_hp_fraction == 1
                and self._estimate_matchup(active, opponent) > 0
            ):
                for move in battle.available_moves:
                    if (
                        move.boosts
                        and sum(move.boosts.values()) >= 2
                        and move.target == "self"
                        and min(
                            [active.boosts[s] for s, v in move.boosts.items() if v > 0]
                        )
                        < 6
                    ):
                        return self.create_order(move)

            move = max(
                battle.available_moves,
                key=lambda m: m.base_power
                * (1.5 if m.type in active.types else 1)
                * (
                    physical_ratio
                    if m.category == MoveCategory.PHYSICAL
                    else special_ratio
                )
                * m.accuracy
                * m.expected_hits
                * opponent.damage_multiplier(m),
            )
            return self.create_order(
                move, dynamax=self._should_dynamax(battle, n_remaining_mons)
            )

        if battle.available_switches:
            switches: List[Pokemon] = battle.available_switches
            return self.create_order(
                max(
                    switches,
                    key=lambda s: self._estimate_matchup(s, opponent),
                )
            )

        return self.choose_random_move(battle)



"""This module defines the Teambuilder abstract class, which represents objects yielding
Pokemon Showdown teams in the context of communicating with Pokemon Showdown.
"""

from abc import ABC, abstractmethod
from typing import List

from poke_env.stats import STATS_TO_IDX
from poke_env.teambuilder.teambuilder_pokemon import TeambuilderPokemon


class Teambuilder(ABC):
    """Teambuilder objects allow the generation of teams by Player instances.

    They must implement the yield_team method, which must return a valid
    packed-formatted showdown team every time it is called.

    This format is a custom format decribed in Pokemon's showdown protocol
    documentation:
    https://github.com/smogon/pokemon-showdown/blob/master/PROTOCOL.md#team-format

    This class also implements a helper function to convert teams from the classical
    showdown team text format into the packed-format.
    """

    @abstractmethod
    def yield_team(self) -> str:
        """Returns a packed-format team."""

    @staticmethod
    def parse_showdown_team(team: str) -> List[TeambuilderPokemon]:
        """Converts a showdown-formatted team string into a list of TeambuilderPokemon
        objects.

        This method can be used when using teams built in the showdown teambuilder.

        :param team: The showdown-format team to convert.
        :type team: str
        :return: The formatted team.
        :rtype: list of TeambuilderPokemon
        """
        current_mon = TeambuilderPokemon()
        current_mon_has_been_added = True
        mons: List[TeambuilderPokemon] = []

        for line in team.split("\n"):
            while line and line.startswith(" "):
                line = line[1:]

            if line == "":
                if not current_mon_has_been_added:
                    mons.append(current_mon)
                current_mon_has_been_added = True
            elif line.startswith("Ability"):
                ability = line.replace("Ability: ", "")
                current_mon.ability = ability.strip()
            elif line.startswith("Level: "):
                level = line.replace("Level: ", "")
                current_mon.level = int(level.strip())
            elif line.startswith("Happiness: "):
                happiness = line.replace("Happiness: ", "")
                current_mon.happiness = int(happiness.strip())
            elif line.startswith("EVs: "):
                evs = line.replace("EVs: ", "")
                split_evs = evs.split(" / ")
                for ev in split_evs:
                    n, stat = ev.split(" ")[:2]
                    idx = STATS_TO_IDX[stat.lower()]
                    current_mon.evs[idx] = int(n)
            elif line.startswith("IVs: "):
                ivs = line.replace("IVs: ", "")
                ivs_split = ivs.split(" / ")
                for iv in ivs_split:
                    n, stat = iv.split(" ")[:2]
                    idx = STATS_TO_IDX[stat.lower()]
                    current_mon.ivs[idx] = int(n)
            elif line.startswith("- "):
                line = line.replace("- ", "").strip()
                current_mon.moves.append(line)
            elif line.startswith("Shiny"):
                current_mon.shiny = line.strip().endswith("Yes")
            elif line.startswith("Gigantamax"):
                current_mon.gmax = line.strip().endswith("Yes")
            elif line.strip().endswith(" Nature"):
                nature = line.strip().replace(" Nature", "")
                current_mon.nature = nature
            elif line.startswith("Hidden Power: "):
                hp_type = line.replace("Hidden Power: ", "").strip()
                current_mon.hiddenpowertype = hp_type
            elif line.startswith("Tera Type: "):
                tera_type = line.replace("Tera Type: ", "").strip()
                current_mon.tera_type = tera_type
            else:
                current_mon = TeambuilderPokemon()
                current_mon_has_been_added = False
                if "@" in line:
                    mon_info, item = line.split(" @ ")
                    current_mon.item = item.strip()
                else:
                    mon_info = line
                split_mon_info = mon_info.split(" ")

                if split_mon_info[-1] == "(M)":
                    current_mon.gender = "M"
                    split_mon_info.pop()
                if split_mon_info[-1] == "(F)":
                    current_mon.gender = "F"
                    split_mon_info.pop()
                if split_mon_info[-1].endswith(")"):
                    for i, info in enumerate(split_mon_info):
                        if info[0] == "(":
                            current_mon.species = " ".join(split_mon_info[i:])[1:-1]
                            split_mon_info = split_mon_info[:i]
                            break
                    current_mon.nickname = " ".join(split_mon_info)
                current_mon.nickname = " ".join(split_mon_info)
        if not current_mon_has_been_added:
            mons.append(current_mon)
        return mons

    @staticmethod
    def join_team(team: List[TeambuilderPokemon]) -> str:
        """Converts a list of TeambuilderPokemon objects into the corresponding packed
        showdown team format.

        :param team: The list of TeambuilderPokemon objects that form the team.
        :type team: list of TeambuilderPokemon
        :return: The formatted team string.
        :rtype: str"""
        return "]".join([mon.formatted for mon in team])



"""This module defines the Effect class, which represents in-game effects."""

from __future__ import annotations

import logging
from enum import Enum, auto, unique
from typing import Dict, Set


@unique
class Effect(Enum):
    """Enumeration, represent an effect a Pokemon can be affected by."""

    UNKNOWN = auto()
    AFTER_YOU = auto()
    AFTERMATH = auto()
    AQUA_RING = auto()
    AROMATHERAPY = auto()
    AROMA_VEIL = auto()
    ATTRACT = auto()
    AUTOTOMIZE = auto()
    BAD_DREAMS = auto()
    BANEFUL_BUNKER = auto()
    BATTLE_BOND = auto()
    BEAK_BLAST = auto()
    BIDE = auto()
    BIND = auto()
    BURNING_BULWARK = auto()
    BURN_UP = auto()
    CELEBRATE = auto()
    CHARGE = auto()
    CLAMP = auto()
    COMMANDER = auto()
    CONFUSION = auto()
    COURT_CHANGE = auto()
    CRAFTY_SHIELD = auto()
    CUD_CHEW = auto()
    CURSE = auto()
    CUSTAP_BERRY = auto()
    DANCER = auto()
    DEFENSE_CURL = auto()
    DESTINY_BOND = auto()
    DISABLE = auto()
    DISGUISE = auto()
    DOOM_DESIRE = auto()
    DRAGON_CHEER = auto()
    DYNAMAX = auto()
    EERIE_SPELL = auto()
    ELECTRIC_TERRAIN = auto()
    ELECTRIFY = auto()
    EMBARGO = auto()
    EMERGENCY_EXIT = auto()
    ENCORE = auto()
    ENDURE = auto()
    FALLEN = auto()
    FALLEN1 = auto()
    FALLEN2 = auto()
    FALLEN3 = auto()
    FALLEN4 = auto()
    FALLEN5 = auto()
    FAIRY_LOCK = auto()
    FEINT = auto()
    FICKLE_BEAM = auto()
    FIRE_SPIN = auto()
    FLASH_FIRE = auto()
    FLINCH = auto()
    FLOWER_VEIL = auto()
    FOCUS_BAND = auto()
    FOCUS_ENERGY = auto()
    FOCUS_PUNCH = auto()
    FOLLOW_ME = auto()
    FORESIGHT = auto()
    FOREWARN = auto()
    FUTURE_SIGHT = auto()
    GASTRO_ACID = auto()
    GLAIVE_RUSH = auto()
    GRAVITY = auto()
    GRUDGE = auto()
    GUARD_DOG = auto()
    GUARD_SPLIT = auto()
    GULP_MISSILE = auto()
    G_MAX_CENTIFERNO = auto()
    G_MAX_CHI_STRIKE = auto()
    G_MAX_ONE_BLOW = auto()
    G_MAX_RAPID_FLOW = auto()
    G_MAX_SANDBLAST = auto()
    HADRON_ENGINE = auto()
    HEAL_BELL = auto()
    HEAL_BLOCK = auto()
    HEALER = auto()
    HELPING_HAND = auto()
    HYDRATION = auto()
    HYPERSPACE_FURY = auto()
    HYPERSPACE_HOLE = auto()
    ICE_FACE = auto()
    ILLUSION = auto()
    IMMUNITY = auto()
    IMPRISON = auto()
    INFESTATION = auto()
    INGRAIN = auto()
    INNARDS_OUT = auto()
    INSOMNIA = auto()
    INSTRUCT = auto()
    IRON_BARBS = auto()
    KINGS_SHIELD = auto()
    LASER_FOCUS = auto()
    LEECH_SEED = auto()
    LEPPA_BERRY = auto()
    LIGHTNING_ROD = auto()
    LIMBER = auto()
    LIQUID_OOZE = auto()
    LOCKED_MOVE = auto()
    LOCK_ON = auto()
    MAGIC_COAT = auto()
    MAGMA_STORM = auto()
    MAGNET_RISE = auto()
    MAGNITUDE = auto()
    MAT_BLOCK = auto()
    MAX_GUARD = auto()
    MIMIC = auto()
    MIMICRY = auto()
    MIND_READER = auto()
    MINIMIZE = auto()
    MIRACLE_EYE = auto()
    MIST = auto()
    MISTY_TERRAIN = auto()
    MUMMY = auto()
    MUST_RECHARGE = auto()
    NEUTRALIZING_GAS = auto()
    NIGHTMARE = auto()
    NO_RETREAT = auto()
    OBLIVIOUS = auto()
    OBSTRUCT = auto()
    OCTOLOCK = auto()
    ORICHALCUM_PULSE = auto()
    OWN_TEMPO = auto()
    PARTIALLY_TRAPPED = auto()
    PASTEL_VEIL = auto()
    PERISH0 = auto()
    PERISH1 = auto()
    PERISH2 = auto()
    PERISH3 = auto()
    PHANTOM_FORCE = auto()
    POLTERGEIST = auto()
    POWDER = auto()
    POWER_CONSTRUCT = auto()
    POWER_SHIFT = auto()
    POWER_SPLIT = auto()
    POWER_TRICK = auto()
    PROTECT = auto()
    PROTECTIVE_PADS = auto()
    PROTOSYNTHESIS = auto()
    PROTOSYNTHESISATK = auto()
    PROTOSYNTHESISDEF = auto()
    PROTOSYNTHESISSPA = auto()
    PROTOSYNTHESISSPD = auto()
    PROTOSYNTHESISSPE = auto()
    PSYCHIC_TERRAIN = auto()
    PURSUIT = auto()
    QUARK_DRIVE = auto()
    QUARKDRIVEATK = auto()
    QUARKDRIVEDEF = auto()
    QUARKDRIVESPA = auto()
    QUARKDRIVESPD = auto()
    QUARKDRIVESPE = auto()
    QUASH = auto()
    QUICK_CLAW = auto()
    QUICK_DRAW = auto()
    QUICK_GUARD = auto()
    RAGE = auto()
    RAGE_POWDER = auto()
    REFLECT = auto()
    RIPEN = auto()
    ROOST = auto()
    ROUGH_SKIN = auto()
    SAFEGUARD = auto()
    SAFETY_GOGGLES = auto()
    SALT_CURE = auto()
    SAND_TOMB = auto()
    SCREEN_CLEANER = auto()
    SHADOW_FORCE = auto()
    SHED_SKIN = auto()
    SILK_TRAP = auto()
    SKETCH = auto()
    SKILL_SWAP = auto()
    SKY_DROP = auto()
    SLOW_START = auto()
    SMACK_DOWN = auto()
    SNAP_TRAP = auto()
    SNATCH = auto()
    SPARKLING_ARIA = auto()
    SPEED_SWAP = auto()
    SPIKY_SHIELD = auto()
    SPITE = auto()
    SPOTLIGHT = auto()
    STICKY_HOLD = auto()
    STICKY_WEB = auto()
    STOCKPILE = auto()
    STOCKPILE1 = auto()
    STOCKPILE2 = auto()
    STOCKPILE3 = auto()
    STORM_DRAIN = auto()
    STRUGGLE = auto()
    SUBSTITUTE = auto()
    SUCTION_CUPS = auto()
    SUPREME_OVERLORD = auto()
    SYRUP_BOMB = auto()
    SWEET_VEIL = auto()
    SYMBIOSIS = auto()
    SYNCHRONIZE = auto()
    TAR_SHOT = auto()
    TAUNT = auto()
    TELEKINESIS = auto()
    TELEPATHY = auto()
    TERA_SHELL = auto()
    TERA_SHIFT = auto()
    TIDY_UP = auto()
    TOXIC_DEBRIS = auto()
    THERMAL_EXCHANGE = auto()
    THROAT_CHOP = auto()
    THUNDER_CAGE = auto()
    TORMENT = auto()
    TRAPPED = auto()
    TRICK = auto()
    TYPEADD = auto()
    TYPECHANGE = auto()
    UPROAR = auto()
    VITAL_SPIRIT = auto()
    WANDERING_SPIRIT = auto()
    WATER_BUBBLE = auto()
    WATER_VEIL = auto()
    WHIRLPOOL = auto()
    WIDE_GUARD = auto()
    WIMP_OUT = auto()
    WRAP = auto()
    YAWN = auto()
    ZERO_TO_HERO = auto()

    def __str__(self) -> str:
        return f"{self.name} (effect) object"

    @staticmethod
    def from_showdown_message(message: str) -> Effect:
        """Returns the Effect object corresponding to the message.

        :param message: The message to convert.
        :type message: str
        :return: The corresponding Effect object.
        :rtype: Effect
        """
        message = message.replace("item: ", "")
        message = message.replace("move: ", "")
        message = message.replace("ability: ", "")
        message = message.replace(" ", "_")
        message = message.replace("-", "_")
        message = message.upper()

        if message == "FALLENUNDEFINED":
            message = "FALLEN"

        try:
            return Effect[message]
        except KeyError:
            logging.getLogger("poke-env").warning(
                "Unexpected effect '%s' received. Effect.UNKNOWN will be used instead. "
                "If this is unexpected, please open an issue at "
                "https://github.com/hsahovic/poke-env/issues/ along with this error "
                "message and a description of your program.",
                message,
            )
            return Effect.UNKNOWN

    @staticmethod
    def from_data(message: str) -> Effect:
        """Returns the Effect object corresponding to the string in static data.

        :param message: The message to convert.
        :type message: str
        :return: The corresponding Effect object.
        :rtype: Effect
        """
        message = message.replace("_", "")
        message = message.replace(" ", "")
        message = message.replace("-", "")
        message = message.upper()
        try:
            return _FROM_DATA[message]
        except KeyError:
            logging.getLogger("poke-env").warning(
                "Unexpected effect '%s' received. Effect.UNKNOWN will be used instead. "
                "If this is unexpected, please open an issue at "
                "https://github.com/hsahovic/poke-env/issues/ along with this error "
                "message and a description of your program.",
                message,
            )
            return Effect.UNKNOWN

    @property
    def breaks_protect(self):
        """
        :return: Whether this effect breaks protect-like states.
        :rtype: bool
        """
        return self in _PROTECT_BREAKING_EFFECTS

    @property
    def ends_on_move(self) -> bool:
        """
        :return: Whether this effect ends when a pokemon moves.
        :rtype: bool
        """
        return self in _ENDS_ON_MOVE_EFFECTS

    @property
    def ends_on_switch(self) -> bool:
        """
        :return: Whether this effect ends when the pokemon switches out.
        :rtype: bool
        """
        return self.is_volatile_status or self in _ENDS_ON_SWITCH_EFFECTS

    @property
    def ends_on_turn(self) -> bool:
        """
        :return: Whether this effect ends at the end of the turn.
        :rtype: bool
        """
        return self in _ENDS_ON_TURN_EFFECTS

    @property
    def is_turn_countable(self) -> bool:
        """
        :return: Whether it is useful to keep track of the number of turns this effect
            has been active for.
        :rtype: bool
        """
        return self in _TURN_COUNTER_EFFECTS

    @property
    def is_action_countable(self) -> bool:
        """
        :return: Whether it is useful to keep track of the number of times this effect
            has been activated.
        :rtype: bool
        """
        return self in _ACTION_COUNTER_EFFECTS

    @property
    def is_volatile_status(self) -> bool:
        """
        :return: Whether the effect is a volatile status, brought by a move
        :rtype: bool
        """
        return self in _VOLATILE_STATUS_EFFECTS

    @property
    def is_from_ability(self) -> bool:
        """
        :return: Whether this effect is a result of an ability
        :rtype: bool
        """
        return self in _FROM_ABILITY_EFFECTS

    @property
    def is_from_item(self) -> bool:
        """
        :return: Whether this effect is a result of an item
        :rtype: bool
        """
        return self in _FROM_ITEM_EFFECTS

    @property
    def is_from_move(self) -> bool:
        """
        :return: Whether this effect is a result of a move
        :rtype: bool
        """
        return self in _FROM_MOVE_EFFECTS


_FROM_ABILITY_EFFECTS: Set[Effect] = {
    Effect.AFTERMATH,
    Effect.AROMA_VEIL,
    Effect.BAD_DREAMS,
    Effect.BATTLE_BOND,
    Effect.COMMANDER,
    Effect.CUD_CHEW,
    Effect.DANCER,
    Effect.DISGUISE,
    Effect.EMERGENCY_EXIT,
    Effect.FLASH_FIRE,
    Effect.FLOWER_VEIL,
    Effect.FOREWARN,
    Effect.GUARD_DOG,
    Effect.GULP_MISSILE,
    Effect.HADRON_ENGINE,
    Effect.HEALER,
    Effect.HYDRATION,
    Effect.ICE_FACE,
    Effect.ILLUSION,
    Effect.IMMUNITY,
    Effect.INNARDS_OUT,
    Effect.IRON_BARBS,
    Effect.LIGHTNING_ROD,
    Effect.LIMBER,
    Effect.LIQUID_OOZE,
    Effect.MIMICRY,
    Effect.MUMMY,
    Effect.NEUTRALIZING_GAS,
    Effect.OBLIVIOUS,
    Effect.ORICHALCUM_PULSE,
    Effect.OWN_TEMPO,
    Effect.PASTEL_VEIL,
    Effect.POWER_CONSTRUCT,
    Effect.PROTECTIVE_PADS,
    Effect.PROTOSYNTHESIS,
    Effect.PROTOSYNTHESISATK,
    Effect.PROTOSYNTHESISDEF,
    Effect.PROTOSYNTHESISSPA,
    Effect.PROTOSYNTHESISSPD,
    Effect.PROTOSYNTHESISSPE,
    Effect.QUARK_DRIVE,
    Effect.QUARKDRIVEATK,
    Effect.QUARKDRIVEDEF,
    Effect.QUARKDRIVESPA,
    Effect.QUARKDRIVESPD,
    Effect.QUARKDRIVESPE,
    Effect.QUICK_DRAW,
    Effect.RIPEN,
    Effect.ROUGH_SKIN,
    Effect.SCREEN_CLEANER,
    Effect.SHED_SKIN,
    Effect.SLOW_START,
    Effect.STICKY_HOLD,
    Effect.STORM_DRAIN,
    Effect.SUCTION_CUPS,
    Effect.SUPREME_OVERLORD,
    Effect.SWEET_VEIL,
    Effect.SYMBIOSIS,
    Effect.SYNCHRONIZE,
    Effect.TELEPATHY,
    Effect.TERA_SHELL,
    Effect.TERA_SHIFT,
    Effect.TOXIC_DEBRIS,
    Effect.THERMAL_EXCHANGE,
    Effect.TRAPPED,
    Effect.VITAL_SPIRIT,
    Effect.WANDERING_SPIRIT,
    Effect.WATER_BUBBLE,
    Effect.WATER_VEIL,
    Effect.WIMP_OUT,
    Effect.ZERO_TO_HERO,
}

_FROM_ITEM_EFFECTS: Set[Effect] = {
    Effect.CUSTAP_BERRY,
    Effect.FOCUS_BAND,
    Effect.HEAL_BELL,
    Effect.LEPPA_BERRY,
    Effect.PROTOSYNTHESISATK,
    Effect.PROTOSYNTHESISDEF,
    Effect.PROTOSYNTHESISSPA,
    Effect.PROTOSYNTHESISSPD,
    Effect.PROTOSYNTHESISSPE,
    Effect.QUARKDRIVEATK,
    Effect.QUARKDRIVEDEF,
    Effect.QUARKDRIVESPA,
    Effect.QUARKDRIVESPD,
    Effect.QUARKDRIVESPE,
    Effect.QUICK_CLAW,
    Effect.SAFETY_GOGGLES,
}

_FROM_MOVE_EFFECTS: Set[Effect] = {
    Effect.AFTER_YOU,
    Effect.AQUA_RING,
    Effect.AROMATHERAPY,
    Effect.ATTRACT,
    Effect.AUTOTOMIZE,
    Effect.BANEFUL_BUNKER,
    Effect.BEAK_BLAST,
    Effect.BIDE,
    Effect.BIND,
    Effect.BURNING_BULWARK,
    Effect.BURN_UP,
    Effect.CELEBRATE,
    Effect.CHARGE,
    Effect.CLAMP,
    Effect.CONFUSION,
    Effect.COURT_CHANGE,
    Effect.CRAFTY_SHIELD,
    Effect.CURSE,
    Effect.DEFENSE_CURL,
    Effect.DESTINY_BOND,
    Effect.DISABLE,
    Effect.DOOM_DESIRE,
    Effect.DRAGON_CHEER,
    Effect.EERIE_SPELL,
    Effect.ELECTRIC_TERRAIN,
    Effect.ELECTRIFY,
    Effect.EMBARGO,
    Effect.ENCORE,
    Effect.ENDURE,
    Effect.FAIRY_LOCK,
    Effect.FEINT,
    Effect.FICKLE_BEAM,
    Effect.FIRE_SPIN,
    Effect.FLINCH,
    Effect.FOCUS_ENERGY,
    Effect.FOCUS_PUNCH,
    Effect.FOLLOW_ME,
    Effect.FORESIGHT,
    Effect.FUTURE_SIGHT,
    Effect.GASTRO_ACID,
    Effect.GLAIVE_RUSH,
    Effect.GRAVITY,
    Effect.GRUDGE,
    Effect.GUARD_SPLIT,
    Effect.G_MAX_CENTIFERNO,
    Effect.G_MAX_CHI_STRIKE,
    Effect.G_MAX_ONE_BLOW,
    Effect.G_MAX_RAPID_FLOW,
    Effect.G_MAX_SANDBLAST,
    Effect.HELPING_HAND,
    Effect.HYPERSPACE_FURY,
    Effect.HYPERSPACE_HOLE,
    Effect.IMPRISON,
    Effect.INFESTATION,
    Effect.INGRAIN,
    Effect.INSOMNIA,
    Effect.INSTRUCT,
    Effect.KINGS_SHIELD,
    Effect.LASER_FOCUS,
    Effect.LEECH_SEED,
    Effect.LOCKED_MOVE,
    Effect.LOCK_ON,
    Effect.MAGIC_COAT,
    Effect.MAGMA_STORM,
    Effect.MAGNET_RISE,
    Effect.MAGNITUDE,
    Effect.MAT_BLOCK,
    Effect.MAX_GUARD,
    Effect.MIMIC,
    Effect.MIND_READER,
    Effect.MINIMIZE,
    Effect.MIRACLE_EYE,
    Effect.MIST,
    Effect.MISTY_TERRAIN,
    Effect.MUST_RECHARGE,
    Effect.NIGHTMARE,
    Effect.NO_RETREAT,
    Effect.OBSTRUCT,
    Effect.OCTOLOCK,
    Effect.PARTIALLY_TRAPPED,
    Effect.PERISH0,
    Effect.PERISH1,
    Effect.PERISH2,
    Effect.PERISH3,
    Effect.PHANTOM_FORCE,
    Effect.POLTERGEIST,
    Effect.POWDER,
    Effect.POWER_SHIFT,
    Effect.POWER_SPLIT,
    Effect.POWER_TRICK,
    Effect.PROTECT,
    Effect.PSYCHIC_TERRAIN,
    Effect.PURSUIT,
    Effect.QUASH,
    Effect.QUICK_GUARD,
    Effect.RAGE,
    Effect.RAGE_POWDER,
    Effect.REFLECT,
    Effect.ROOST,
    Effect.SAFEGUARD,
    Effect.SALT_CURE,
    Effect.SAND_TOMB,
    Effect.SHADOW_FORCE,
    Effect.SILK_TRAP,
    Effect.SKETCH,
    Effect.SKILL_SWAP,
    Effect.SKY_DROP,
    Effect.SMACK_DOWN,
    Effect.SNAP_TRAP,
    Effect.SNATCH,
    Effect.SPARKLING_ARIA,
    Effect.SPEED_SWAP,
    Effect.SPIKY_SHIELD,
    Effect.SPITE,
    Effect.SPOTLIGHT,
    Effect.STICKY_WEB,
    Effect.STOCKPILE,
    Effect.STOCKPILE1,
    Effect.STOCKPILE2,
    Effect.STOCKPILE3,
    Effect.STRUGGLE,
    Effect.SUBSTITUTE,
    Effect.SYRUP_BOMB,
    Effect.TAR_SHOT,
    Effect.TAUNT,
    Effect.TELEKINESIS,
    Effect.TIDY_UP,
    Effect.THROAT_CHOP,
    Effect.THUNDER_CAGE,
    Effect.TRAPPED,
    Effect.TRICK,
    Effect.TYPEADD,
    Effect.TYPECHANGE,
    Effect.UPROAR,
    Effect.WHIRLPOOL,
    Effect.WIDE_GUARD,
    Effect.WRAP,
    Effect.YAWN,
}

_VOLATILE_STATUS_EFFECTS: Set[Effect] = {
    Effect.FORESIGHT,
    Effect.DISABLE,
    Effect.SMACK_DOWN,
    Effect.CURSE,
    Effect.NIGHTMARE,
    Effect.TAR_SHOT,
    Effect.LOCKED_MOVE,
    Effect.BIDE,
    Effect.SNATCH,
    Effect.ENCORE,
    Effect.SPOTLIGHT,
    Effect.LASER_FOCUS,
    Effect.ELECTRIFY,
    Effect.SILK_TRAP,
    Effect.HEAL_BLOCK,
    Effect.MIRACLE_EYE,
    Effect.ATTRACT,
    Effect.DESTINY_BOND,
    Effect.RAGE,
    Effect.AQUA_RING,
    Effect.TAUNT,
    Effect.TELEKINESIS,
    Effect.GRUDGE,
    Effect.SYRUP_BOMB,
    Effect.YAWN,
    Effect.IMPRISON,
    Effect.MAGIC_COAT,
    Effect.STOCKPILE,
    Effect.DEFENSE_CURL,
    Effect.EMBARGO,
    Effect.FOLLOW_ME,
    Effect.GLAIVE_RUSH,
    Effect.CONFUSION,
    Effect.INGRAIN,
    Effect.BANEFUL_BUNKER,
    Effect.PROTECT,
    Effect.GASTRO_ACID,
    Effect.POWER_TRICK,
    Effect.TORMENT,
    Effect.BURNING_BULWARK,
    Effect.DRAGON_CHEER,
    Effect.SPARKLING_ARIA,
    Effect.HELPING_HAND,
    Effect.FLINCH,
    Effect.OCTOLOCK,
    Effect.PARTIALLY_TRAPPED,
    Effect.KINGS_SHIELD,
    Effect.MINIMIZE,
    Effect.NO_RETREAT,
    Effect.CHARGE,
    Effect.MUST_RECHARGE,
    Effect.MAGNET_RISE,
    Effect.MAX_GUARD,
    Effect.POWER_SHIFT,
    Effect.SUBSTITUTE,
    Effect.RAGE_POWDER,
    Effect.ROOST,
    Effect.FOCUS_ENERGY,
    Effect.SALT_CURE,
    Effect.LEECH_SEED,
    Effect.POWDER,
    Effect.SPIKY_SHIELD,
    Effect.ENDURE,
    Effect.UPROAR,
    Effect.OBSTRUCT,
}

_PROTECT_BREAKING_EFFECTS: Set[Effect] = {
    Effect.FEINT,
    Effect.G_MAX_ONE_BLOW,
    Effect.G_MAX_RAPID_FLOW,
    Effect.SHADOW_FORCE,
    Effect.PHANTOM_FORCE,
    Effect.HYPERSPACE_FURY,
    Effect.HYPERSPACE_HOLE,
}

_TURN_COUNTER_EFFECTS: Set[Effect] = {
    Effect.BIDE,
    Effect.BIND,
    Effect.CLAMP,
    Effect.DISABLE,
    Effect.DOOM_DESIRE,
    Effect.DYNAMAX,
    Effect.EMBARGO,
    Effect.ENCORE,
    Effect.FIRE_SPIN,
    Effect.FUTURE_SIGHT,
    Effect.GRAVITY,
    Effect.G_MAX_CENTIFERNO,
    Effect.G_MAX_SANDBLAST,
    Effect.HEAL_BLOCK,
    Effect.INFESTATION,
    Effect.MAGMA_STORM,
    Effect.MAGNET_RISE,
    Effect.SAND_TOMB,
    Effect.SKY_DROP,
    Effect.SLOW_START,
    Effect.SNAP_TRAP,
    Effect.TAUNT,
    Effect.TELEKINESIS,
    Effect.THROAT_CHOP,
    Effect.THUNDER_CAGE,
    Effect.UPROAR,
    Effect.WHIRLPOOL,
    Effect.WRAP,
}

_ENDS_ON_MOVE_EFFECTS = {
    Effect.GLAIVE_RUSH,
    Effect.CHARGE,
    Effect.DANCER,
    Effect.GRUDGE,
    Effect.DESTINY_BOND,
    Effect.RAGE,
    Effect.INSTRUCT,
    Effect.FOCUS_PUNCH,
}

_ENDS_ON_SWITCH_EFFECTS = {
    Effect.MIND_READER,
    Effect.MUMMY,
    Effect.SPEED_SWAP,
    Effect.TYPECHANGE,
    Effect.SKILL_SWAP,
    Effect.PERISH0,
    Effect.PERISH1,
    Effect.PERISH2,
    Effect.PERISH3,
    Effect.FLASH_FIRE,
}

_ENDS_ON_TURN_EFFECTS = {
    Effect.AFTER_YOU,
    Effect.BANEFUL_BUNKER,
    Effect.BEAK_BLAST,
    Effect.BURNING_BULWARK,
    Effect.CRAFTY_SHIELD,
    Effect.FEINT,
    Effect.FLINCH,
    Effect.FOCUS_PUNCH,
    Effect.FOLLOW_ME,
    Effect.KINGS_SHIELD,
    Effect.CUSTAP_BERRY,
    Effect.MIND_READER,
    Effect.MAGIC_COAT,
    Effect.OBSTRUCT,
    Effect.PROTECT,
    Effect.QUASH,
    Effect.QUICK_CLAW,
    Effect.QUICK_DRAW,
    Effect.QUICK_GUARD,
    Effect.RAGE_POWDER,
    Effect.ROOST,
    Effect.SPIKY_SHIELD,
    Effect.SPOTLIGHT,
    Effect.WIDE_GUARD,
    Effect.INSTRUCT,
    Effect.FOCUS_PUNCH,
}

_ACTION_COUNTER_EFFECTS: Set[Effect] = {Effect.RAGE, Effect.STOCKPILE}

_FROM_DATA: Dict[str, Effect] = {
    "UNKNOWN": Effect.UNKNOWN,
    "AFTERYOU": Effect.AFTER_YOU,
    "AFTERMATH": Effect.AFTERMATH,
    "AQUARING": Effect.AQUA_RING,
    "AROMATHERAPY": Effect.AROMATHERAPY,
    "AROMAVEIL": Effect.AROMA_VEIL,
    "ATTRACT": Effect.ATTRACT,
    "AUTOTOMIZE": Effect.AUTOTOMIZE,
    "BADDREAMS": Effect.BAD_DREAMS,
    "BANEFULBUNKER": Effect.BANEFUL_BUNKER,
    "BATTLEBOND": Effect.BATTLE_BOND,
    "BIDE": Effect.BIDE,
    "BIND": Effect.BIND,
    "BURNINGBULWARK": Effect.BURNING_BULWARK,
    "BURNUP": Effect.BURN_UP,
    "CELEBRATE": Effect.CELEBRATE,
    "CHARGE": Effect.CHARGE,
    "CLAMP": Effect.CLAMP,
    "COMMANDER": Effect.COMMANDER,
    "CONFUSION": Effect.CONFUSION,
    "COURTCHANGE": Effect.COURT_CHANGE,
    "CRAFTYSHIELD": Effect.CRAFTY_SHIELD,
    "CUDCHEW": Effect.CUD_CHEW,
    "CURSE": Effect.CURSE,
    "CUSTAPBERRY": Effect.CUSTAP_BERRY,
    "DANCER": Effect.DANCER,
    "DEFENSECURL": Effect.DEFENSE_CURL,
    "DESTINYBOND": Effect.DESTINY_BOND,
    "DISABLE": Effect.DISABLE,
    "DISGUISE": Effect.DISGUISE,
    "DOOMDESIRE": Effect.DOOM_DESIRE,
    "DRAGONCHEER": Effect.DRAGON_CHEER,
    "DYNAMAX": Effect.DYNAMAX,
    "EERIESPELL": Effect.EERIE_SPELL,
    "ELECTRICTERRAIN": Effect.ELECTRIC_TERRAIN,
    "ELECTRIFY": Effect.ELECTRIFY,
    "EMBARGO": Effect.EMBARGO,
    "EMERGENCYEXIT": Effect.EMERGENCY_EXIT,
    "ENCORE": Effect.ENCORE,
    "ENDURE": Effect.ENDURE,
    "FALLEN": Effect.FALLEN,
    "FALLEN1": Effect.FALLEN1,
    "FALLEN2": Effect.FALLEN2,
    "FALLEN3": Effect.FALLEN3,
    "FALLEN4": Effect.FALLEN4,
    "FALLEN5": Effect.FALLEN5,
    "FAIRYLOCK": Effect.FAIRY_LOCK,
    "FEINT": Effect.FEINT,
    "FICKLEBEAM": Effect.FICKLE_BEAM,
    "FIRESPIN": Effect.FIRE_SPIN,
    "FLASHFIRE": Effect.FLASH_FIRE,
    "FLINCH": Effect.FLINCH,
    "FLOWERVEIL": Effect.FLOWER_VEIL,
    "FOCUSBAND": Effect.FOCUS_BAND,
    "FOCUSENERGY": Effect.FOCUS_ENERGY,
    "FOCUSPUNCH": Effect.FOCUS_PUNCH,
    "FOLLOWME": Effect.FOLLOW_ME,
    "FORESIGHT": Effect.FORESIGHT,
    "FOREWARN": Effect.FOREWARN,
    "FUTURESIGHT": Effect.FUTURE_SIGHT,
    "GASTROACID": Effect.GASTRO_ACID,
    "GUARDDOG": Effect.GUARD_DOG,
    "GLAIVERUSH": Effect.GLAIVE_RUSH,
    "GRAVITY": Effect.GRAVITY,
    "GRUDGE": Effect.GRUDGE,
    "GUARDSPLIT": Effect.GUARD_SPLIT,
    "GULPMISSILE": Effect.GULP_MISSILE,
    "GMAXCENTIFERNO": Effect.G_MAX_CENTIFERNO,
    "GMAXCHISTRIKE": Effect.G_MAX_CHI_STRIKE,
    "GMAXONEBLOW": Effect.G_MAX_ONE_BLOW,
    "GMAXRAPIDFLOW": Effect.G_MAX_RAPID_FLOW,
    "GMAXSANDBLAST": Effect.G_MAX_SANDBLAST,
    "HADRONENGINE": Effect.HADRON_ENGINE,
    "HEALBELL": Effect.HEAL_BELL,
    "HEALBLOCK": Effect.HEAL_BLOCK,
    "HEALER": Effect.HEALER,
    "HELPINGHAND": Effect.HELPING_HAND,
    "HYDRATION": Effect.HYDRATION,
    "HYPERSPACEFURY": Effect.HYPERSPACE_FURY,
    "HYPERSPACEHOLE": Effect.HYPERSPACE_HOLE,
    "ICEFACE": Effect.ICE_FACE,
    "ILLUSION": Effect.ILLUSION,
    "IMMUNITY": Effect.IMMUNITY,
    "IMPRISON": Effect.IMPRISON,
    "INFESTATION": Effect.INFESTATION,
    "INGRAIN": Effect.INGRAIN,
    "INNARDSOUT": Effect.INNARDS_OUT,
    "INSTRUCT": Effect.INSTRUCT,
    "INSOMNIA": Effect.INSOMNIA,
    "IRONBARBS": Effect.IRON_BARBS,
    "KINGSSHIELD": Effect.KINGS_SHIELD,
    "LASERFOCUS": Effect.LASER_FOCUS,
    "LEECHSEED": Effect.LEECH_SEED,
    "LEPPABERRY": Effect.LEPPA_BERRY,
    "LIGHTNINGROD": Effect.LIGHTNING_ROD,
    "LIMBER": Effect.LIMBER,
    "LIQUIDOOZE": Effect.LIQUID_OOZE,
    "LOCKEDMOVE": Effect.LOCKED_MOVE,
    "LOCKON": Effect.LOCK_ON,
    "MAGICCOAT": Effect.MAGIC_COAT,
    "MAGMASTORM": Effect.MAGMA_STORM,
    "MAGNETRISE": Effect.MAGNET_RISE,
    "MAGNITUDE": Effect.MAGNITUDE,
    "MATBLOCK": Effect.MAT_BLOCK,
    "MAXGUARD": Effect.MAX_GUARD,
    "MIMIC": Effect.MIMIC,
    "MIMICRY": Effect.MIMICRY,
    "MINDREADER": Effect.MIND_READER,
    "MINIMIZE": Effect.MINIMIZE,
    "MIRACLEEYE": Effect.MIRACLE_EYE,
    "MIST": Effect.MIST,
    "MISTYTERRAIN": Effect.MISTY_TERRAIN,
    "MUMMY": Effect.MUMMY,
    "MUSTRECHARGE": Effect.MUST_RECHARGE,
    "NEUTRALIZINGGAS": Effect.NEUTRALIZING_GAS,
    "NIGHTMARE": Effect.NIGHTMARE,
    "NORETREAT": Effect.NO_RETREAT,
    "OBLIVIOUS": Effect.OBLIVIOUS,
    "OBSTRUCT": Effect.OBSTRUCT,
    "OCTOLOCK": Effect.OCTOLOCK,
    "ORICHALCUMPULSE": Effect.ORICHALCUM_PULSE,
    "OWNTEMPO": Effect.OWN_TEMPO,
    "PARTIALLYTRAPPED": Effect.PARTIALLY_TRAPPED,
    "PASTELVEIL": Effect.PASTEL_VEIL,
    "PERISH0": Effect.PERISH0,
    "PERISH1": Effect.PERISH1,
    "PERISH2": Effect.PERISH2,
    "PERISH3": Effect.PERISH3,
    "PHANTOMFORCE": Effect.PHANTOM_FORCE,
    "POLTERGEIST": Effect.POLTERGEIST,
    "POWDER": Effect.POWDER,
    "POWERCONSTRUCT": Effect.POWER_CONSTRUCT,
    "POWERSHIFT": Effect.POWER_SHIFT,
    "POWERSPLIT": Effect.POWER_SPLIT,
    "POWERTRICK": Effect.POWER_TRICK,
    "PROTECT": Effect.PROTECT,
    "PROTECTIVEPADS": Effect.PROTECTIVE_PADS,
    "PROTOSYNTHESIS": Effect.PROTOSYNTHESIS,
    "PROTOSYNTHESISATK": Effect.PROTOSYNTHESISATK,
    "PROTOSYNTHESISDEF": Effect.PROTOSYNTHESISDEF,
    "PROTOSYNTHESISSPA": Effect.PROTOSYNTHESISSPA,
    "PROTOSYNTHESISSPD": Effect.PROTOSYNTHESISSPD,
    "PROTOSYNTHESISSPE": Effect.PROTOSYNTHESISSPE,
    "PSYCHICTERRAIN": Effect.PSYCHIC_TERRAIN,
    "PURSUIT": Effect.PURSUIT,
    "QUARKDRIVE": Effect.QUARK_DRIVE,
    "QUARKDRIVEATK": Effect.QUARKDRIVEATK,
    "QUARKDRIVEDEF": Effect.QUARKDRIVEDEF,
    "QUARKDRIVESPA": Effect.QUARKDRIVESPA,
    "QUARKDRIVESPD": Effect.QUARKDRIVESPD,
    "QUARKDRIVESPE": Effect.QUARKDRIVESPE,
    "QUASH": Effect.QUASH,
    "QUICKCLAW": Effect.QUICK_CLAW,
    "QUICKDRAW": Effect.QUICK_DRAW,
    "QUICKGUARD": Effect.QUICK_GUARD,
    "RAGE": Effect.RAGE,
    "RAGEPOWDER": Effect.RAGE_POWDER,
    "REFLECT": Effect.REFLECT,
    "RIPEN": Effect.RIPEN,
    "ROOST": Effect.ROOST,
    "ROUGHSKIN": Effect.ROUGH_SKIN,
    "SAFEGUARD": Effect.SAFEGUARD,
    "SAFETYGOGGLES": Effect.SAFETY_GOGGLES,
    "SALTCURE": Effect.SALT_CURE,
    "SANDTOMB": Effect.SAND_TOMB,
    "SCREENCLEANER": Effect.SCREEN_CLEANER,
    "SHADOWFORCE": Effect.SHADOW_FORCE,
    "SHEDSKIN": Effect.SHED_SKIN,
    "SILKTRAP": Effect.SILK_TRAP,
    "SKETCH": Effect.SKETCH,
    "SKILLSWAP": Effect.SKILL_SWAP,
    "SKYDROP": Effect.SKY_DROP,
    "SLOWSTART": Effect.SLOW_START,
    "SMACKDOWN": Effect.SMACK_DOWN,
    "SNAPTRAP": Effect.SNAP_TRAP,
    "SNATCH": Effect.SNATCH,
    "SPARKLINGARIA": Effect.SPARKLING_ARIA,
    "SPEEDSWAP": Effect.SPEED_SWAP,
    "SPIKYSHIELD": Effect.SPIKY_SHIELD,
    "SPITE": Effect.SPITE,
    "SPOTLIGHT": Effect.SPOTLIGHT,
    "STICKY_HOLD": Effect.STICKY_HOLD,
    "STICKY_WEB": Effect.STICKY_WEB,
    "STOCKPILE": Effect.STOCKPILE,
    "STOCKPILE1": Effect.STOCKPILE1,
    "STOCKPILE2": Effect.STOCKPILE2,
    "STOCKPILE3": Effect.STOCKPILE3,
    "STORMDRAIN": Effect.STORM_DRAIN,
    "STRUGGLE": Effect.STRUGGLE,
    "SUBSTITUTE": Effect.SUBSTITUTE,
    "SUCTIONCUPS": Effect.SUCTION_CUPS,
    "SUPREMEOVERLORD": Effect.SUPREME_OVERLORD,
    "SYRUPBOMB": Effect.SYRUP_BOMB,
    "SWEETVEIL": Effect.SWEET_VEIL,
    "SYMBIOSIS": Effect.SYMBIOSIS,
    "SYNCHRONIZE": Effect.SYNCHRONIZE,
    "TARSHOT": Effect.TAR_SHOT,
    "TAUNT": Effect.TAUNT,
    "TELEKINESIS": Effect.TELEKINESIS,
    "TELEPATHY": Effect.TELEPATHY,
    "TERASHELL": Effect.TERA_SHELL,
    "TERASHIFT": Effect.TERA_SHIFT,
    "TIDYUP": Effect.TIDY_UP,
    "TOXICDEBRIS": Effect.TOXIC_DEBRIS,
    "THERMALEXCHANGE": Effect.THERMAL_EXCHANGE,
    "THROATCHOP": Effect.THROAT_CHOP,
    "THUNDERCAGE": Effect.THUNDER_CAGE,
    "TORMENT": Effect.TORMENT,
    "TRAPPED": Effect.TRAPPED,
    "TRICK": Effect.TRICK,
    "TYPEADD": Effect.TYPEADD,
    "TYPECHANGE": Effect.TYPECHANGE,
    "UPROAR": Effect.UPROAR,
    "VITALSPIRIT": Effect.VITAL_SPIRIT,
    "WANDERINGSPIRIT": Effect.WANDERING_SPIRIT,
    "WATERBUBBLE": Effect.WATER_BUBBLE,
    "WATERVEIL": Effect.WATER_VEIL,
    "WHIRLPOOL": Effect.WHIRLPOOL,
    "WIDEGUARD": Effect.WIDE_GUARD,
    "WIMPOUT": Effect.WIMP_OUT,
    "WRAP": Effect.WRAP,
    "YAWN": Effect.YAWN,
    "ZERO_TO_HERO": Effect.ZERO_TO_HERO,
}



from logging import Logger
from typing import Any, Dict, List, Optional, Union

from poke_env.environment.abstract_battle import AbstractBattle
from poke_env.environment.move import Move
from poke_env.environment.pokemon import Pokemon
from poke_env.environment.pokemon_type import PokemonType


class Battle(AbstractBattle):
    def __init__(
        self,
        battle_tag: str,
        username: str,
        logger: Logger,
        gen: int,
        save_replays: Union[str, bool] = False,
    ):
        super(Battle, self).__init__(battle_tag, username, logger, save_replays, gen)

        # Turn choice attributes
        self._available_moves: List[Move] = []
        self._available_switches: List[Pokemon] = []
        self._can_dynamax: bool = False
        self._can_mega_evolve: bool = False
        self._can_tera: Optional[PokemonType] = None
        self._can_z_move: bool = False
        self._opponent_can_dynamax = True
        self._opponent_can_mega_evolve = True
        self._opponent_can_z_move = True
        self._opponent_can_tera: bool = False
        self._force_switch: bool = False
        self._maybe_trapped: bool = False
        self._trapped: bool = False

    def clear_all_boosts(self):
        if self.active_pokemon is not None:
            self.active_pokemon.clear_boosts()
        if self.opponent_active_pokemon is not None:
            self.opponent_active_pokemon.clear_boosts()

    def end_illusion(self, pokemon_name: str, details: str):
        if pokemon_name[:2] == self._player_role:
            active = self.active_pokemon
        else:
            active = self.opponent_active_pokemon

        if active is None:
            raise ValueError("Cannot end illusion without an active pokemon.")

        self._end_illusion_on(
            illusioned=active, illusionist=pokemon_name, details=details
        )

    def parse_request(self, request: Dict[str, Any]) -> None:
        """
        Update the object from a request.
        The player's pokemon are all updated, as well as available moves, switches and
        other related information (z move, mega evolution, forced switch...).

        :param request: Parsed JSON request object.
        :type request: dict
        """
        if "wait" in request and request["wait"]:
            self._wait = True
        else:
            self._wait = False

        side = request["side"]

        self._available_moves = []
        self._available_switches = []
        self._can_mega_evolve = False
        self._can_z_move = False
        self._can_dynamax = False
        self._can_tera = None
        self._maybe_trapped = False
        self._reviving = any(
            [m["reviving"] for m in side.get("pokemon", []) if "reviving" in m]
        )
        self._trapped = False
        self._force_switch = request.get("forceSwitch", [False])[0]

        if self._force_switch:
            self._move_on_next_request = True

        self._last_request = request

        if request.get("teamPreview", False):
            self._teampreview = True
            number_of_mons = len(request["side"]["pokemon"])
            self._max_team_size = request.get("maxTeamSize", number_of_mons)
        else:
            self._teampreview = False
        self._update_team_from_request(request["side"])

        if "active" in request:
            active_request = request["active"][0]

            if active_request.get("trapped"):
                self._trapped = True

            if self.active_pokemon is not None:
                self._available_moves.extend(
                    self.active_pokemon.available_moves_from_request(active_request)
                )
            if active_request.get("canMegaEvo", False):
                self._can_mega_evolve = True
            if active_request.get("canZMove", False):
                self._can_z_move = True
            if active_request.get("canDynamax", False):
                self._can_dynamax = True
            if active_request.get("maybeTrapped", False):
                self._maybe_trapped = True
            if active_request.get("canTerastallize", False):
                self._can_tera = PokemonType.from_name(
                    active_request["canTerastallize"]
                )

        if side["pokemon"]:
            self._player_role = side["pokemon"][0]["ident"][:2]

        if not self.trapped and not self.reviving:
            for pokemon in side["pokemon"]:
                if pokemon:
                    pokemon = self._team[pokemon["ident"]]
                    if not pokemon.active and not pokemon.fainted:
                        self._available_switches.append(pokemon)

        if not self.trapped and self.reviving:
            for pokemon in side["pokemon"]:
                if pokemon and pokemon.get("reviving", False):
                    pokemon = self._team[pokemon["ident"]]
                    if not pokemon.active:
                        self._available_switches.append(pokemon)

    def switch(self, pokemon_str: str, details: str, hp_status: str):
        identifier = pokemon_str.split(":")[0][:2]

        if identifier == self._player_role:
            if self.active_pokemon:
                self.active_pokemon.switch_out()
        else:
            if self.opponent_active_pokemon:
                self.opponent_active_pokemon.switch_out()

        pokemon = self.get_pokemon(pokemon_str, details=details)

        pokemon.switch_in(details=details)
        pokemon.set_hp_status(hp_status)

    @property
    def active_pokemon(self) -> Optional[Pokemon]:
        """
        :return: The active pokemon
        :rtype: Optional[Pokemon]
        """
        for pokemon in self.team.values():
            if pokemon.active:
                return pokemon
        return None

    @property
    def all_active_pokemons(self) -> List[Optional[Pokemon]]:
        """
        :return: A list containing all active pokemons and/or Nones.
        :rtype: List[Optional[Pokemon]]
        """
        return [self.active_pokemon, self.opponent_active_pokemon]

    @property
    def available_moves(self) -> List[Move]:
        """
        :return: The list of moves the player can use during the current move request.
        :rtype: List[Move]
        """
        return self._available_moves

    @property
    def available_switches(self) -> List[Pokemon]:
        """
        :return: The list of switches the player can do during the current move request.
        :rtype: List[Pokemon]
        """
        return self._available_switches

    @property
    def can_dynamax(self) -> bool:
        """
        :return: Whether or not the current active pokemon can dynamax
        :rtype: bool
        """
        return self._can_dynamax

    @property
    def can_mega_evolve(self) -> bool:
        """
        :return: Whether or not the current active pokemon can mega evolve.
        :rtype: bool
        """
        return self._can_mega_evolve

    @property
    def can_tera(self) -> Optional[PokemonType]:
        """
        :return: None, or the type the active pokemon can terastallize into.
        :rtype: PokemonType, optional
        """
        return self._can_tera

    @property
    def can_z_move(self) -> bool:
        """
        :return: Whether or not the current active pokemon can z-move.
        :rtype: bool
        """
        return self._can_z_move

    @property
    def force_switch(self) -> bool:
        """
        :return: A boolean indicating whether the active pokemon is forced to switch
            out.
        :rtype: Optional[bool]
        """
        return self._force_switch

    @property
    def grounded(self) -> bool:
        """
        :return: A boolean indicating whether the active pokemon is grounded
        :rtype: bool
        """
        return self.is_grounded(self.active_pokemon) if self.active_pokemon else True

    @property
    def maybe_trapped(self) -> bool:
        """
        :return: A boolean indicating whether the active pokemon is maybe trapped by the
            opponent.
        :rtype: bool
        """
        return self._maybe_trapped

    @property
    def opponent_active_pokemon(self) -> Optional[Pokemon]:
        """
        :return: The opponent active pokemon
        :rtype: Pokemon
        """
        for pokemon in self.opponent_team.values():
            if pokemon.active:
                return pokemon
        return None

    @property
    def opponent_can_dynamax(self) -> bool:
        """
        :return: Whether or not opponent's current active pokemon can dynamax
        :rtype: bool
        """
        return self._opponent_can_dynamax

    @opponent_can_dynamax.setter
    def opponent_can_dynamax(self, value: bool):
        self._opponent_can_dynamax = value

    @property
    def opponent_can_mega_evolve(self) -> Union[bool, List[bool]]:
        """
        :return: Whether or not opponent's current active pokemon can mega-evolve
        :rtype: bool
        """
        return self._opponent_can_mega_evolve

    @opponent_can_mega_evolve.setter
    def opponent_can_mega_evolve(self, value: bool):
        self._opponent_can_mega_evolve = value

    @property
    def opponent_can_tera(self) -> bool:
        """
        :return: Whether or not opponent's current active pokemon can terastallize
        :rtype: bool
        """
        return self._opponent_can_tera

    @property
    def opponent_can_z_move(self) -> Union[bool, List[bool]]:
        """
        :return: Whether or not opponent's current active pokemon can z-move
        :rtype: bool
        """
        return self._opponent_can_z_move

    @opponent_can_z_move.setter
    def opponent_can_z_move(self, value: bool):
        self._opponent_can_z_move = value

    @property
    def trapped(self) -> bool:
        """
        :return: A boolean indicating whether the active pokemon is trapped, either by
            the opponent or as a side effect of one your moves.
        :rtype: bool
        """
        return self._trapped

    @trapped.setter
    def trapped(self, value: bool):
        self._trapped = value
